# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from math import isclose
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import torch

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.schedulers import PRUNING_SCHEDULERS
from nncf.common.pruning.schedulers import PruningScheduler
from nncf.common.pruning.shape_pruning_processor import ShapePruningProcessor
from nncf.common.pruning.statistics import FilterPruningStatistics
from nncf.common.pruning.statistics import PrunedLayerSummary
from nncf.common.pruning.statistics import PrunedModelStatistics
from nncf.common.pruning.statistics import PrunedModelTheoreticalBorderline
from nncf.common.pruning.utils import get_prunable_layers_in_out_channels
from nncf.common.pruning.utils import get_rounded_pruned_element_number
from nncf.common.pruning.weights_flops_calculator import WeightsFlopsCalculator
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import copy_model
from nncf.common.utils.debug import is_debug
from nncf.common.utils.os import safe_open
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.operator_metatypes import PTModuleConv1dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv3dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConvTranspose1dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConvTranspose2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConvTranspose3dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv1dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv2dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv3dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleLinearMetatype
from nncf.torch.layers import NNCF_PRUNING_MODULES_DICT
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.pruning.base_algo import BasePruningAlgoBuilder
from nncf.torch.pruning.base_algo import BasePruningAlgoController
from nncf.torch.pruning.filter_pruning.functions import FILTER_IMPORTANCE_FUNCTIONS
from nncf.torch.pruning.filter_pruning.functions import calculate_binary_mask
from nncf.torch.pruning.filter_pruning.functions import tensor_l2_normalizer
from nncf.torch.pruning.filter_pruning.global_ranking.legr import LeGR
from nncf.torch.pruning.filter_pruning.layers import FilterPruningMask
from nncf.torch.pruning.operations import PT_PRUNING_OPERATOR_METATYPES
from nncf.torch.pruning.operations import ModelPruner
from nncf.torch.pruning.operations import PrunType
from nncf.torch.pruning.operations import PTElementwisePruningOp
from nncf.torch.pruning.structs import PrunedModuleInfo
from nncf.torch.pruning.tensor_processor import PTNNCFPruningTensorProcessor
from nncf.torch.pruning.utils import collect_output_shapes
from nncf.torch.pruning.utils import init_output_masks_in_graph
from nncf.torch.structures import DistributedCallbacksArgs
from nncf.torch.structures import LeGRInitArgs
from nncf.torch.utils import get_filters_num

GENERAL_CONV_LAYER_METATYPES = [
    PTModuleConv1dMetatype,
    PTModuleDepthwiseConv1dSubtype,
    PTModuleConv2dMetatype,
    PTModuleDepthwiseConv2dSubtype,
    PTModuleConv3dMetatype,
    PTModuleDepthwiseConv3dSubtype,
    PTModuleConvTranspose1dMetatype,
    PTModuleConvTranspose2dMetatype,
    PTModuleConvTranspose3dMetatype,
]
LINEAR_LAYER_METATYPES = [PTModuleLinearMetatype]


@PT_COMPRESSION_ALGORITHMS.register("filter_pruning")
class FilterPruningBuilder(BasePruningAlgoBuilder):
    def create_weight_pruning_operation(self, module, node_name):
        return FilterPruningMask(
            module.weight.size(module.target_weight_dim_for_compression),
            node_name,
            module.target_weight_dim_for_compression,
        )

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return FilterPruningController(
            model, self._prunable_types, self.pruned_module_groups_info, self._pruned_norms_operators, self.config
        )

    def _is_pruned_module(self, module) -> bool:
        # Currently prune only Convolutions
        return isinstance(module, tuple(NNCF_PRUNING_MODULES_DICT.keys()))

    def get_op_types_of_pruned_modules(self) -> List[str]:
        types = [v.op_func_name for v in NNCF_PRUNING_MODULES_DICT]
        return types

    def get_types_of_grouping_ops(self) -> List[str]:
        return PTElementwisePruningOp.get_all_op_aliases()


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("pt_filter_pruning")
class FilterPruningController(BasePruningAlgoController):
    def __init__(
        self,
        target_model: NNCFNetwork,
        prunable_types: List[str],
        pruned_module_groups: Clusterization[PrunedModuleInfo],
        pruned_norms_operators: List[Tuple[NNCFNode, FilterPruningMask, torch.nn.Module]],
        config: NNCFConfig,
    ):
        super().__init__(target_model, prunable_types, pruned_module_groups, config)
        params = self.pruning_config.get("params", {})
        self._pruned_norms_operators = pruned_norms_operators
        self.frozen = False
        self._pruning_level = 0
        self.pruning_init = self.pruning_config.get("pruning_init", 0)
        self.pruning_quota = 0.9
        self.normalize_weights = True

        self._graph = self._model.nncf.get_original_graph()
        self._weights_flops_calc = WeightsFlopsCalculator(
            conv_op_metatypes=GENERAL_CONV_LAYER_METATYPES, linear_op_metatypes=LINEAR_LAYER_METATYPES
        )

        self._shape_pruning_proc = ShapePruningProcessor(
            pruning_operations_metatype=PT_PRUNING_OPERATOR_METATYPES, prunable_types=prunable_types
        )
        self.pruning_quotas = {}
        self.nodes_flops: Dict[NNCFNodeName, int] = {}
        self.nodes_params_num: Dict[NNCFNodeName, int] = {}
        self._next_nodes: Dict[int, List[NNCFNodeName]] = {}
        self._output_shapes: Dict[NNCFNodeName, int] = {}
        _, modules_out_channels = get_prunable_layers_in_out_channels(self._graph)
        self._init_pruned_layers_params(modules_out_channels)
        self.flops_count_init()

        self.full_flops = sum(self.nodes_flops.values())
        self.current_flops = self.full_flops
        self.full_params_num = sum(self.nodes_params_num.values())
        self.current_params_num = self.full_params_num
        self.full_filters_num = self._weights_flops_calc.count_filters_num(self._graph, modules_out_channels)
        self.current_filters_num = self.full_filters_num
        self._pruned_layers_num = len(self.pruned_module_groups_info.get_all_nodes())
        self._prunable_layers_num = len(self._model.nncf.get_graph().get_nodes_by_types(self._prunable_types))
        (
            self._min_possible_flops,
            self._min_possible_params,
        ) = self._calculate_flops_and_weights_in_uniformly_pruned_model(1.0)

        self.weights_normalizer = tensor_l2_normalizer  # for all weights in common case
        self.filter_importance = FILTER_IMPORTANCE_FUNCTIONS.get(params.get("filter_importance", "L2"))
        self.ranking_type = params.get("interlayer_ranking_type", "unweighted_ranking")
        self.all_weights = params.get("all_weights", False)
        scheduler_cls = PRUNING_SCHEDULERS.get(params.get("schedule", "exponential"))
        self._scheduler = scheduler_cls(self, params)

        if self.ranking_type == "learned_ranking":
            # In case of learned_ranking ranking type weights shouldn't be normalized
            self.normalize_weights = False
            if params.get("load_ranking_coeffs_path"):
                coeffs_path = params.get("load_ranking_coeffs_path")
                nncf_logger.info(f"Loading ranking coefficients from file {coeffs_path}")
                try:
                    with safe_open(Path(coeffs_path), "r", encoding="utf8") as coeffs_file:
                        loaded_coeffs = json.load(coeffs_file)
                except (ValueError, FileNotFoundError) as err:
                    raise Exception(
                        "Can't load json with ranking coefficients. Please, check format of json file "
                        "and path to the file."
                    ) from err
                ranking_coeffs = {key: tuple(loaded_coeffs[key]) for key in loaded_coeffs}
                nncf_logger.debug(f"Loaded ranking coefficients = {ranking_coeffs}")
                self.ranking_coeffs = ranking_coeffs
            else:
                # Ranking can't be trained without registered init struct LeGRInitArgs
                if not config.has_extra_struct(LeGRInitArgs):
                    raise Exception("Please, register LeGRInitArgs via register_default_init_args function.")
                # Wrapping model for parallelization
                distributed_wrapping_init_args = config.get_extra_struct(DistributedCallbacksArgs)
                target_model = distributed_wrapping_init_args.wrap_model(target_model)
                legr_init_args = config.get_extra_struct(LeGRInitArgs)
                legr_params = params.get("legr_params", {})
                if "max_pruning" not in legr_params:
                    legr_params["max_pruning"] = self._scheduler.target_level
                self.legr = LeGR(self, target_model, legr_init_args, **legr_params)
                self.ranking_coeffs = self.legr.train_global_ranking()
                nncf_logger.debug(f"Trained ranking coefficients = {self.ranking_coeffs}")
                # Unwrapping parallelized model
                _ = distributed_wrapping_init_args.unwrap_model(target_model)
        else:
            self.ranking_coeffs = {node.node_name: (1, 0) for node in self.pruned_module_groups_info.get_all_nodes()}

        # Saving ranking coefficients to the specified file
        if params.get("save_ranking_coeffs_path"):
            nncf_logger.info(f'Saving ranking coefficients to the file {params.get("save_ranking_coeffs_path")}')
            with safe_open(Path(params.get("save_ranking_coeffs_path")), "w", encoding="utf8") as f:
                json.dump(self.ranking_coeffs, f)

        self.set_pruning_level(self.pruning_init)
        self._bn_adaptation = None

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> PruningScheduler:
        return self._scheduler

    @staticmethod
    def get_mask(minfo: PrunedModuleInfo) -> torch.Tensor:
        return minfo.operand.binary_filter_pruning_mask

    @staticmethod
    def set_mask(minfo: PrunedModuleInfo, mask: torch.Tensor) -> None:
        minfo.operand.binary_filter_pruning_mask = mask

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        if not quickly_collected_only and is_debug():
            stats = PrunedModelTheoreticalBorderline(
                self._pruned_layers_num,
                self._prunable_layers_num,
                self._min_possible_flops,
                self._min_possible_params,
                self.full_flops,
                self.full_params_num,
            )

            nncf_logger.debug(stats.to_str())

        pruned_layers_summary = {}
        for minfo in self.pruned_module_groups_info.get_all_nodes():
            layer_name = str(minfo.module_scope)
            if layer_name not in pruned_layers_summary:
                pruned_layers_summary[layer_name] = PrunedLayerSummary(
                    layer_name,
                    list(minfo.module.weight.size()),
                    list(self.mask_shape(minfo)),
                    self.pruning_level_for_mask(minfo),
                )

        self._update_benchmark_statistics()
        model_statistics = PrunedModelStatistics(
            self.full_flops,
            self.current_flops,
            self.full_params_num,
            self.current_params_num,
            self.full_filters_num,
            self.current_filters_num,
            list(pruned_layers_summary.values()),
        )

        stats = FilterPruningStatistics(
            model_statistics, self.scheduler.current_pruning_level, self.scheduler.target_level, self.prune_flops
        )

        nncf_stats = NNCFStatistics()
        nncf_stats.register("filter_pruning", stats)
        return nncf_stats

    @property
    def pruning_level(self) -> float:
        """Global pruning level in the model"""
        return self._pruning_level

    def freeze(self, freeze: bool = True):
        self.frozen = freeze

    def _init_pruned_layers_params(self, full_out_channels):
        # 1. Collect nodes output shapes
        self._output_shapes = collect_output_shapes(self._graph)

        # 2. Init next_nodes for every pruning cluster
        self._next_nodes = self._shape_pruning_proc.get_next_nodes(self._graph, self.pruned_module_groups_info)

        # 3. Init pruning quotas
        for cluster in self.pruned_module_groups_info.get_all_clusters():
            self.pruning_quotas[cluster.id] = np.floor(
                full_out_channels[cluster.elements[0].node_name] * self.pruning_quota
            )

    def flops_count_init(self) -> None:
        self.nodes_flops, self.nodes_params_num = self._weights_flops_calc.count_flops_and_weights_per_node(
            self._graph, self._output_shapes
        )

    def _calculate_flops_and_weights_in_uniformly_pruned_model(self, pruning_level: float) -> Tuple[int, int]:
        """
        Prune all prunable modules in model by pruning_level level and returns number of weights and
        flops of the pruned model.

        :param pruning_level: proportion of zero filters in all modules
        :return: flops number in pruned model
        """
        (
            tmp_in_channels,
            tmp_out_channels,
        ) = self._shape_pruning_proc.calculate_in_out_channels_in_uniformly_pruned_model(
            graph=self._graph,
            pruning_groups=self.pruned_module_groups_info,
            pruning_groups_next_nodes=self._next_nodes,
            pruning_level=pruning_level,
        )

        return self._weights_flops_calc.count_flops_and_weights(
            graph=self._graph,
            output_shapes=self._output_shapes,
            input_channels=tmp_in_channels,
            output_channels=tmp_out_channels,
        )

    def _find_uniform_pruning_level_for_target_flops(self, target_flops_pruning_level: float) -> float:
        """
        Searching for the minimal uniform layer-wise weight pruning level (proportion of zero filters in a layer)
         needed to achieve the target pruning level in flops.

        :param target_flops_pruning_level: target proportion of flops that should be pruned in the model
        :return: uniform pruning level for all layers
        """
        error = 0.01
        target_flops = self.full_flops * (1 - target_flops_pruning_level)
        left, right = 0.0, 1.0
        while abs(right - left) > error:
            middle = (left + right) / 2
            flops, params_num = self._calculate_flops_and_weights_in_uniformly_pruned_model(middle)
            if flops < target_flops:
                right = middle
            else:
                left = middle
        flops, params_num = self._calculate_flops_and_weights_in_uniformly_pruned_model(right)
        if flops <= target_flops:
            self.current_flops = flops
            self.current_params_num = params_num
            return right
        raise nncf.InternalError(
            "Can't prune the model to get the required "
            "pruning level in flops = {}".format(target_flops_pruning_level)
        )

    def set_pruning_level(
        self, pruning_level: Union[float, Dict[int, float]], run_batchnorm_adaptation: bool = False
    ) -> None:
        """
        Set the global or groupwise pruning level in the model.
        If pruning_level is a float, the correspoding global pruning level is set in the model,
        either in terms of the percentage of filters pruned or as the percentage of flops
        removed, the latter being true in case the "prune_flops" flag of the controller is
        set to True.
        If pruning_level is a dict, the keys should correspond to layer group id's and the
        values to groupwise pruning level to be set in the model.
        """
        groupwise_pruning_levels_set = isinstance(pruning_level, dict)
        passed_pruning_level = pruning_level

        if not self.frozen:
            nncf_logger.info("Computing filter importance scores and binary masks...")
            with torch.no_grad():
                if self.all_weights:
                    if groupwise_pruning_levels_set:
                        raise nncf.InternalError("Cannot set group-wise pruning levels with all_weights=True")
                    # Non-uniform (global) importance-score-based pruning according
                    # to the global pruning level
                    if self.prune_flops:
                        self._set_binary_masks_for_pruned_modules_globally_by_flops_target(pruning_level)
                    else:
                        self._set_binary_masks_for_pruned_modules_globally(pruning_level)
                else:
                    if groupwise_pruning_levels_set:
                        group_ids = [group.id for group in self.pruned_module_groups_info.get_all_clusters()]
                        if set(pruning_level.keys()) != set(group_ids):
                            raise nncf.InternalError(
                                "Groupwise pruning level dict keys do not correspond to layer group ids"
                            )
                    else:
                        # Pruning uniformly with the same pruning level across layers
                        if self.prune_flops:
                            # Looking for layerwise pruning level needed for the required flops pruning level
                            pruning_level = self._find_uniform_pruning_level_for_target_flops(pruning_level)
                    self._set_binary_masks_for_pruned_modules_groupwise(pruning_level)

        self._propagate_masks()
        if not groupwise_pruning_levels_set:
            self._pruning_level = passed_pruning_level
        else:
            self._pruning_level = self._calculate_global_weight_pruning_level()

        if run_batchnorm_adaptation:
            self._run_batchnorm_adaptation()

    def _calculate_global_weight_pruning_level(self) -> float:
        full_param_count = 0
        pruned_param_count = 0
        for minfo in self.pruned_module_groups_info.get_all_nodes():
            layer_param_count = sum(p.numel() for p in minfo.module.parameters() if p.requires_grad)
            layer_weight_pruning_level = self.pruning_level_for_mask(minfo)
            full_param_count += layer_param_count
            pruned_param_count += layer_param_count * layer_weight_pruning_level
        return pruned_param_count / full_param_count

    @property
    def current_groupwise_pruning_level(self) -> Dict[int, float]:
        """
        Return the dict of layer group id's and corresponding current groupwise
        pruning levels in the model
        """
        groupwise_pruning_level_dict = {}
        for group in self.pruned_module_groups_info.get_all_clusters():
            groupwise_pruning_level_dict[group.id] = self.pruning_level_for_mask(group.elements[0])
        return groupwise_pruning_level_dict

    def _set_binary_masks_for_pruned_modules_groupwise(self, pruning_level: Union[float, Dict[int, float]]) -> None:
        """
        Set the binary mask values according to groupwise pruning level.
        If pruning_level is a float, set the pruning level uniformly across groups.
        If pruning_level is a dict, set specific pruning levels corresponding to each group.
        """
        nncf_logger.debug("Updating binary masks for pruned modules.")
        groupwise_pruning_levels_set = isinstance(pruning_level, dict)

        for group in self.pruned_module_groups_info.get_all_clusters():
            group_pruning_level = pruning_level[group.id] if groupwise_pruning_levels_set else pruning_level

            filters_num = torch.tensor([get_filters_num(minfo.module) for minfo in group.elements])
            assert torch.all(filters_num == filters_num[0])
            device = group.elements[0].module.weight.device

            cumulative_filters_importance = torch.zeros(filters_num[0]).to(device)
            # 1. Calculate cumulative importance for all filters in group
            for minfo in group.elements:
                filters_importance = self.filter_importance(
                    minfo.module.weight, minfo.module.target_weight_dim_for_compression
                )
                cumulative_filters_importance += filters_importance

            # 2. Calculate threshold
            num_of_sparse_elems = get_rounded_pruned_element_number(
                cumulative_filters_importance.size(0), group_pruning_level
            )
            if num_of_sparse_elems == 0:
                nncf_logger.warning("Binary masks are identity matrix, please check your config.")
            threshold = sorted(cumulative_filters_importance)[min(num_of_sparse_elems, filters_num[0] - 1)]
            mask = calculate_binary_mask(cumulative_filters_importance, threshold)

            # 3. Set binary masks for filter
            for minfo in group.elements:
                pruning_module = minfo.operand
                pruning_module.binary_filter_pruning_mask = mask

        # Calculate actual flops and weights number with new masks
        self._update_benchmark_statistics()

    def _set_binary_masks_for_pruned_modules_globally(self, pruning_level: float) -> None:
        """
        Set the binary mask values for layer groups according to the global pruning level.
        Filter importance scores in each group are merged into a single global list and a
        threshold value separating the pruning_level proportion of the least important filters
        in the model is calculated. Filters are pruned globally according to the threshold value.
        """
        nncf_logger.debug("Setting new binary masks for all pruned modules together.")
        filter_importances = []
        # 1. Calculate importances for all groups of  filters
        for group in self.pruned_module_groups_info.get_all_clusters():
            filters_num = torch.tensor([get_filters_num(minfo.module) for minfo in group.elements])
            assert torch.all(filters_num == filters_num[0])
            device = group.elements[0].module.weight.device

            cumulative_filters_importance = torch.zeros(filters_num[0]).to(device)
            # Calculate cumulative importance for all filters in this group
            for minfo in group.elements:
                normalized_weight = self.weights_normalizer(minfo.module.weight)
                filters_importance = self.filter_importance(
                    normalized_weight, minfo.module.target_weight_dim_for_compression
                )
                cumulative_filters_importance += filters_importance

            filter_importances.append(cumulative_filters_importance)

        # 2. Calculate one threshold for all weights
        importances = torch.cat(filter_importances)
        threshold = sorted(importances)[int(pruning_level * importances.size(0))]

        # 3. Set binary masks for filters in groups
        for i, group in enumerate(self.pruned_module_groups_info.get_all_clusters()):
            mask = calculate_binary_mask(filter_importances[i], threshold)
            for minfo in group.elements:
                pruning_module = minfo.operand
                pruning_module.binary_filter_pruning_mask = mask

        # Calculate actual flops and weights number with new masks
        self._update_benchmark_statistics()

    def _set_binary_masks_for_pruned_modules_globally_by_flops_target(self, target_flops_pruning_level: float) -> None:
        """
        Sorting all prunable filters in the network by importance and pruning the amount of the
        least important filters sufficient to achieve the target pruning level by flops.
        Filters are pruned one-by-one and the corresponding flops value is checked.

        :param target_flops_pruning_level: target proportion of flops removed from the model
        :return:
        """
        target_flops = self.full_flops * (1 - target_flops_pruning_level)

        # 1. Initialize masks
        for minfo in self.pruned_module_groups_info.get_all_nodes():
            new_mask = torch.ones(get_filters_num(minfo.module)).to(minfo.module.weight.device)
            self.set_mask(minfo, new_mask)

        # 2. Calculate filter importances for all prunable groups
        filter_importances = []
        cluster_indexes = []
        filter_indexes = []

        for cluster in self.pruned_module_groups_info.get_all_clusters():
            filters_num = torch.tensor([get_filters_num(minfo.module) for minfo in cluster.elements])
            assert torch.all(filters_num == filters_num[0])
            device = cluster.elements[0].module.weight.device

            cumulative_filters_importance = torch.zeros(filters_num[0]).to(device)
            # Calculate cumulative importance for all filters in this group
            for minfo in cluster.elements:
                weight = minfo.module.weight
                if self.normalize_weights:
                    weight = self.weights_normalizer(weight)
                filters_importance = self.filter_importance(weight, minfo.module.target_weight_dim_for_compression)
                scaled_importance = (
                    self.ranking_coeffs[minfo.node_name][0] * filters_importance
                    + self.ranking_coeffs[minfo.node_name][1]
                )
                cumulative_filters_importance += scaled_importance

            filter_importances.append(cumulative_filters_importance)
            cluster_indexes.append(cluster.id * torch.ones_like(cumulative_filters_importance))
            filter_indexes.append(torch.arange(len(cumulative_filters_importance)))

        importances = torch.cat(filter_importances)
        cluster_indexes = torch.cat(cluster_indexes)
        filter_indexes = torch.cat(filter_indexes)

        # 3. Sort all filter groups by importances and prune the least important filters
        # until target flops pruning level is achieved
        sorted_importances = sorted(zip(importances, cluster_indexes, filter_indexes), key=lambda x: x[0])
        cur_num = 0
        tmp_in_channels, tmp_out_channels = get_prunable_layers_in_out_channels(self._graph)
        tmp_pruning_quotas = self.pruning_quotas.copy()

        while cur_num < len(sorted_importances):
            cluster_idx = int(sorted_importances[cur_num][1])
            filter_idx = int(sorted_importances[cur_num][2])

            if tmp_pruning_quotas[cluster_idx] > 0:
                tmp_pruning_quotas[cluster_idx] -= 1
            else:
                cur_num += 1
                continue

            cluster = self.pruned_module_groups_info.get_cluster_by_id(cluster_idx)
            self._shape_pruning_proc.prune_cluster_shapes(
                cluster=cluster,
                pruned_elems=1,
                pruning_groups_next_nodes=self._next_nodes,
                input_channels=tmp_in_channels,
                output_channels=tmp_out_channels,
            )

            for node in cluster.elements:
                node.operand.binary_filter_pruning_mask[filter_idx] = 0

            flops, params_num = self._weights_flops_calc.count_flops_and_weights(
                graph=self._graph,
                output_shapes=self._output_shapes,
                input_channels=tmp_in_channels,
                output_channels=tmp_out_channels,
            )
            if flops < target_flops:
                self.current_flops = flops
                self.current_params_num = params_num
                return
            cur_num += 1
        raise nncf.InternalError("Can't prune model to asked flops pruning level")

    def _propagate_masks(self):
        nncf_logger.debug("Propagating pruning masks")
        # 1. Propagate masks for all modules
        graph = self.model.nncf.get_original_graph()

        init_output_masks_in_graph(graph, self.pruned_module_groups_info.get_all_nodes())
        MaskPropagationAlgorithm(graph, PT_PRUNING_OPERATOR_METATYPES, PTNNCFPruningTensorProcessor).mask_propagation()

        # 2. Set the masks for Batch/Group Norms
        pruned_node_modules = []
        for node, pruning_block, node_module in self._pruned_norms_operators:
            if node_module not in pruned_node_modules:
                # Setting masks for BN nodes
                pruning_block.binary_filter_pruning_mask = node.attributes["output_mask"].tensor
                pruned_node_modules.append(node_module)

    def prepare_for_export(self):
        """
        Applies pruning masks to layer weights before exporting the model to ONNX.
        """
        self._propagate_masks()

        pruned_layers_stats = self.get_stats_for_pruned_modules()
        nncf_logger.info(f"Pruned layers statistics: \n{pruned_layers_stats}")

    def compression_stage(self) -> CompressionStage:
        target_pruning_level = self.scheduler.target_level
        actual_pruning_level = self.compression_rate
        if actual_pruning_level == 0:
            return CompressionStage.UNCOMPRESSED
        if (
            isclose(actual_pruning_level, target_pruning_level, abs_tol=1e-5)
            or actual_pruning_level > target_pruning_level
        ):
            return CompressionStage.FULLY_COMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    @property
    def compression_rate(self):
        if self.prune_flops:
            return 1 - self.current_flops / self.full_flops
        return self.pruning_level

    @compression_rate.setter
    def compression_rate(self, compression_rate):
        is_pruning_controller_frozen = self.frozen
        self.freeze(False)
        self.set_pruning_level(compression_rate)
        self.freeze(is_pruning_controller_frozen)

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()
        self._scheduler.current_pruning_level = 0.0

    @property
    def maximal_compression_rate(self) -> float:
        if self.prune_flops:
            return 1 - self._min_possible_flops / max(self.full_flops, 1)
        return 1.0

    def _calculate_num_of_sparse_elements_by_node(self) -> Dict[str, int]:
        num_of_sparse_elements_by_node = {}
        for minfo in self.pruned_module_groups_info.get_all_nodes():
            mask = self.get_mask(minfo)
            num_of_sparse_elements_by_node[minfo.node_name] = mask.view(-1).size(0) - mask.nonzero().size(0)
        return num_of_sparse_elements_by_node

    def _update_benchmark_statistics(self):
        tmp_in_channels, tmp_out_channels = self._shape_pruning_proc.calculate_in_out_channels_by_masks(
            graph=self._graph,
            pruning_groups=self.pruned_module_groups_info,
            pruning_groups_next_nodes=self._next_nodes,
            num_of_sparse_elements_by_node=self._calculate_num_of_sparse_elements_by_node(),
        )

        self.current_filters_num = self._weights_flops_calc.count_filters_num(
            graph=self._graph, output_channels=tmp_out_channels
        )

        self.current_flops, self.current_params_num = self._weights_flops_calc.count_flops_and_weights(
            graph=self._graph,
            output_shapes=self._output_shapes,
            input_channels=tmp_in_channels,
            output_channels=tmp_out_channels,
        )

    def _run_batchnorm_adaptation(self):
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                **extract_bn_adaptation_init_params(self.config, "filter_pruning")
            )
        self._bn_adaptation.run(self.model)

    def strip_model(self, model: NNCFNetwork, do_copy: bool = False) -> NNCFNetwork:
        if do_copy:
            model = copy_model(model)

        graph = model.nncf.get_original_graph()
        ModelPruner(model, graph, PT_PRUNING_OPERATOR_METATYPES, PrunType.FILL_ZEROS).prune_model()

        return model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass
from dataclasses import field
from typing import Dict, List, Optional, Set, Tuple

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api


@api(canonical_alias="nncf.Subgraph")
@dataclass
class Subgraph:
    """
    Defines the ignored subgraph as follows: A subgraph comprises all nodes along
    all simple paths in the graph from input to output nodes.

    :param inputs: Input node names.
    :type inputs: List[str]
    :param outputs: Output node names.
    :type outputs: List[str]
    """

    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)


def get_ignored_node_names_from_subgraph(graph: NNCFGraph, subgraph: Subgraph) -> List[str]:
    """
    Returns all names that should be ignored according to given subgraph.

    :param graph: Given NNCFGraph.
    :param subgraph: Given subgraph instance.
    :return: All names that should be ignored according to given subgraph.
    """
    ignored_names = set()
    for start_node_name in subgraph.inputs:
        for end_node_name in subgraph.outputs:
            if start_node_name == end_node_name:
                # For networkx<3.3 nx.get_all_simple_paths returns empty path for this case
                node = graph.get_node_by_name(start_node_name)
                ignored_names.add(node.node_name)
                continue
            for path in graph.get_all_simple_paths(start_node_name, end_node_name):
                for node_key in path:
                    node = graph.get_node_by_key(node_key)
                    ignored_names.add(node.node_name)

    return list(sorted(ignored_names))


@api(canonical_alias="nncf.IgnoredScope")
@dataclass
class IgnoredScope:
    """
    Provides an option to specify portions of model to be excluded from compression.

    The ignored scope defines model sub-graphs that should be excluded from the compression process such as
    quantization, pruning and etc.

    Example:

    ..  code-block:: python

            import nncf

            # Exclude by node name:
            node_names = ['node_1', 'node_2', 'node_3']
            ignored_scope = nncf.IgnoredScope(names=node_names)

            # Exclude using regular expressions:
            patterns = ['node_\\d']
            ignored_scope = nncf.IgnoredScope(patterns=patterns)

            # Exclude by operation type:

            # OpenVINO opset https://docs.openvino.ai/latest/openvino_docs_ops_opset.html
            operation_types = ['Multiply', 'GroupConvolution', 'Interpolate']
            ignored_scope = nncf.IgnoredScope(types=operation_types)

            # ONNX opset https://github.com/onnx/onnx/blob/main/docs/Operators.md
            operation_types = ['Mul', 'Conv', 'Resize']
            ignored_scope = nncf.IgnoredScope(types=operation_types)

    **Note:** Operation types must be specified according to the model framework.

    :param names: List of ignored node names.
    :type names: List[str]
    :param patterns: List of regular expressions that define patterns for names of ignored nodes.
    :type patterns: List[str]
    :param types: List of ignored operation types.
    :type types: List[str]
    :param subgraphs: List of ignored subgraphs.
    :type subgraphs: List[Subgraph]
    :param validate: If set to True, then a RuntimeError will be raised if any ignored scope does not match
      in the model graph.
    :type types: bool
    """

    names: List[str] = field(default_factory=list)
    patterns: List[str] = field(default_factory=list)
    types: List[str] = field(default_factory=list)
    subgraphs: List[Subgraph] = field(default_factory=list)
    validate: bool = True


def get_difference_ignored_scope(ignored_scope_1: IgnoredScope, ignored_scope_2: IgnoredScope) -> IgnoredScope:
    """
    Returns ignored scope with rules from 'ignored_scope_1' not presented at 'ignored_scope_2'

    :param ignored_scope_1: First ignored scope.
    :param ignored_scope_2: Second ignored scope.
    :return: Ignored scope.
    """
    return IgnoredScope(
        names=list(set(ignored_scope_1.names) - set(ignored_scope_2.names)),
        patterns=list(set(ignored_scope_1.patterns) - set(ignored_scope_2.patterns)),
        types=list(set(ignored_scope_1.types) - set(ignored_scope_2.types)),
        subgraphs=[subgraph for subgraph in ignored_scope_1.subgraphs if subgraph not in ignored_scope_2.subgraphs],
        validate=ignored_scope_1.validate,
    )


def convert_ignored_scope_to_list(ignored_scope: Optional[IgnoredScope]) -> List[str]:
    """
    Convert the contents of the `IgnoredScope` class to the legacy ignored
    scope format.

    :param ignored_scope: The ignored scope.
    :return: An ignored scope in the legacy format as list.
    """
    results: List[str] = []
    if ignored_scope is None:
        return results
    results.extend(ignored_scope.names)
    for p in ignored_scope.patterns:
        results.append("{re}" + p)
    if ignored_scope.types:
        raise nncf.InternalError("Legacy ignored scope format does not support operation types")
    return results


def get_matched_ignored_scope_info(
    ignored_scope: IgnoredScope, nncf_graphs: List[NNCFGraph]
) -> Tuple[IgnoredScope, Dict[str, Set[str]]]:
    """
    Returns matched ignored scope for provided graphs along with all found matches.
    The resulted ignored scope consist of all matched rules.
    The found matches consist of a dictionary with a rule name as a key and matched node names as a value.

    :param ignored_scope: Ignored scope instance.
    :param nncf_graphs: Graphs.
    :returns: Matched ignored scope along with all matches.
    """
    names, patterns, types, subgraphs_numbers = set(), set(), set(), set()  # type: ignore
    matches = {"names": names, "patterns": set(), "types": set(), "subgraphs": set()}

    for graph in nncf_graphs:
        if ignored_scope.names or ignored_scope.patterns:
            node_names = set(node.node_name for node in graph.nodes.values())

            for ignored_node_name in filter(lambda name: name in node_names, ignored_scope.names):
                names.add(ignored_node_name)

            for str_pattern in ignored_scope.patterns:
                pattern = re.compile(str_pattern)
                pattern_matched_names = set(filter(pattern.match, node_names))
                if pattern_matched_names:
                    matches["patterns"].update(pattern_matched_names)
                    patterns.add(str_pattern)

        for node in graph.get_nodes_by_types(ignored_scope.types):
            matches["types"].add(node.node_name)
            types.add(node.node_type)

        for i, subgraph in enumerate(ignored_scope.subgraphs):
            names_from_subgraph = get_ignored_node_names_from_subgraph(graph, subgraph)
            if names_from_subgraph:
                matches["subgraphs"].update(names_from_subgraph)
                subgraphs_numbers.add(i)

    matched_ignored_scope = IgnoredScope(
        names=list(names),
        patterns=list(patterns),
        types=list(types),
        subgraphs=[subgraph for i, subgraph in enumerate(ignored_scope.subgraphs) if i in subgraphs_numbers],
        validate=ignored_scope.validate,
    )
    return matched_ignored_scope, matches


def _info_matched_ignored_scope(matches: Dict[str, Set[str]]) -> None:
    """
    Log matches.

    :param matches: Matches.
    """
    for rule_type, rules in matches.items():
        if rules:
            nncf_logger.info(f"{len(rules)} ignored nodes were found by {rule_type} in the NNCFGraph")


def _error_unmatched_ignored_scope(unmatched_ignored_scope: IgnoredScope) -> str:
    """
    Returns an error message for unmatched ignored scope.

    :param unmatched_ignored_scope: Unmatched ignored scope.
    :return str: Error message.
    """
    err_msg = "\n"
    for ignored_type in ("names", "types", "patterns"):
        unmatched_rules = getattr(unmatched_ignored_scope, ignored_type)
        if unmatched_rules:
            err_msg += f"Ignored nodes that matches {ignored_type} {unmatched_rules} were not found in the NNCFGraph.\n"
    for subgraph in unmatched_ignored_scope.subgraphs:
        err_msg += (
            f"Ignored nodes that matches subgraph with input names {subgraph.inputs} "
            f"and output names {subgraph.outputs} were not found in the NNCFGraph.\n"
        )
    return err_msg


def _check_ignored_scope_strictly_matched(ignored_scope: IgnoredScope, matched_ignored_scope: IgnoredScope) -> None:
    """
    Passes when ignored_scope and matched_ignored_scope are equal, otherwise - raises ValidationError.

    :param ignored_scope: Ignored scope.
    :param matched_ignored_scope: Matched ignored scope.
    """
    unmatched_ignored_scope = get_difference_ignored_scope(ignored_scope, matched_ignored_scope)
    if (
        unmatched_ignored_scope.names
        or unmatched_ignored_scope.types
        or unmatched_ignored_scope.patterns
        or unmatched_ignored_scope.subgraphs
    ):
        raise nncf.ValidationError(_error_unmatched_ignored_scope(unmatched_ignored_scope))


def get_ignored_node_names_from_ignored_scope(
    ignored_scope: IgnoredScope, nncf_graph: NNCFGraph, strict: bool = True
) -> Set[str]:
    """
    Returns ignored names according to ignored scope and NNCFGraph.
    If strict is True, raises nncf.ValidationError if any ignored rule was not matched.
    If strict is False, returns all possible matches.

    :param ignored_scope: Ignored scope.
    :param nncf_graph: Graph.
    :param strict: Whether all ignored_scopes must match at least one node or not.
    :return: NNCF node names from given NNCFGraph specified in given ignored scope.
    """
    matched_ignored_scope, matches = get_matched_ignored_scope_info(ignored_scope, [nncf_graph])
    if strict:
        _check_ignored_scope_strictly_matched(ignored_scope, matched_ignored_scope)
    _info_matched_ignored_scope(matches)
    return {name for match in matches.values() for name in match}


def validate_ignored_scope(ignored_scope: IgnoredScope, nncf_graphs: List[NNCFGraph]) -> None:
    """
    Passes whether all rules at 'ignored_scope' have matches at provided graphs, otherwise - raises ValidationError.

    :param ignored_scope: Ignored scope.
    :param nncf_graphs: Graphs.
    """
    matched_ignored_scope, _ = get_matched_ignored_scope_info(ignored_scope, nncf_graphs)
    _check_ignored_scope_strictly_matched(ignored_scope, matched_ignored_scope)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, List, Optional, Tuple

import torch
import torch.fx

import nncf
import nncf.errors
import nncf.tensor
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.experimental.common.tensor_statistics.collectors import MeanReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import ShapeReducer
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.model_transformer import FXModelTransformer
from nncf.experimental.torch.fx.node_utils import get_graph_node_by_name
from nncf.experimental.torch.fx.node_utils import get_tensor_constant_from_node
from nncf.experimental.torch.fx.transformations import constant_update_transformation_builder
from nncf.experimental.torch.fx.transformations import module_insertion_transformation_builder
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.lora_correction import LoraCorrectionAlgorithm
from nncf.quantization.algorithms.weight_compression.torch_backend import PTWeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.weight_lowering import compress_weight
from nncf.tensor import Tensor
from nncf.tensor.definitions import TensorDataType
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import get_const_node
from nncf.torch.model_graph_manager import get_weight_tensor_port_ids
from nncf.torch.quantization.layers import INT4AsymmetricWeightsDecompressor
from nncf.torch.quantization.layers import INT4SymmetricWeightsDecompressor
from nncf.torch.quantization.layers import INT8AsymmetricWeightsDecompressor
from nncf.torch.quantization.layers import INT8SymmetricWeightsDecompressor


class FXWeightCompressionAlgoBackend(WeightCompressionAlgoBackend):
    MATMUL_METATYPES = PTWeightCompressionAlgoBackend.MATMUL_METATYPES
    EMBEDDING_METATYPES = PTWeightCompressionAlgoBackend.EMBEDDING_METATYPES
    CONVOLUTION_METATYPES = PTWeightCompressionAlgoBackend.CONVOLUTION_METATYPES

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return FXWeightCompressionAlgoBackend.MATMUL_METATYPES

    @property
    def embedding_metatypes(self) -> List[OperatorMetatype]:
        return FXWeightCompressionAlgoBackend.EMBEDDING_METATYPES

    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return FXWeightCompressionAlgoBackend.CONVOLUTION_METATYPES

    @staticmethod
    def is_node_with_weights(node: NNCFNode, graph: NNCFGraph) -> bool:
        return PTWeightCompressionAlgoBackend.is_node_with_weights(node, graph)

    @staticmethod
    def get_weight_names_and_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Tuple[str, int]]:
        port_ids = get_weight_tensor_port_ids(node, graph)
        weight_name_port_ids = [(get_const_node(node, pid, graph).node_name, pid) for pid in port_ids]
        return weight_name_port_ids

    @staticmethod
    def get_reduction_axes(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Optional[Tuple[int]]:
        weight_node = get_const_node(node_with_weight, weight_port_id, graph)
        edge = graph.get_edge(weight_node, graph.get_next_nodes(weight_node)[0])

        ndims = len(edge.tensor_shape)
        reduction_axes = None
        if node_with_weight.metatype == om.PTAtenEmbeddingMetatype:
            reduction_axes = [1]
        elif node_with_weight.metatype == om.PTLinearMetatype:
            reduction_axes = [ndims - 1]
        elif node_with_weight.metatype == om.PTMatMulMetatype:
            if weight_port_id == 0:
                reduction_axes = [ndims - 1]
            elif weight_port_id == 1:
                reduction_axes = [max(0, ndims - 2)]
        elif node_with_weight.metatype == om.PTAddmmMetatype:
            if weight_port_id == 1:
                reduction_axes = [ndims - 1]
            elif weight_port_id == 2:
                reduction_axes = [max(0, ndims - 2)]
        elif node_with_weight.metatype in FXWeightCompressionAlgoBackend.CONVOLUTION_METATYPES:
            channel_idx = (
                1
                if node_with_weight.metatype
                in [om.PTConvTranspose1dMetatype, om.PTConvTranspose2dMetatype, om.PTConvTranspose3dMetatype]
                else 0
            )
            reduction_axes = [i for i in range(ndims) if i != channel_idx]
        return tuple(reduction_axes)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        return PTWeightCompressionAlgoBackend.target_point(target_type, target_node_name, port_id)

    def mean_statistic_collector(
        self, reduction_axes: Tuple[int], subset_size: Optional[int] = None
    ) -> TensorCollector:
        mean_reducer = MeanReducer(reduction_axes)
        shape_reducer = ShapeReducer()
        collector = TensorCollector(WCTensorStatistic)
        collector.register_statistic_branch(WCTensorStatistic.MEAN_STAT, mean_reducer, NoopAggregator(subset_size))
        collector.register_statistic_branch(WCTensorStatistic.SHAPE_STAT, shape_reducer, NoopAggregator(subset_size))
        return collector

    @staticmethod
    def get_activation_port_id(node: NNCFNode, graph: NNCFGraph) -> int:
        return PTWeightCompressionAlgoBackend.get_activation_port_id(node, graph)

    def get_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: torch.fx.GraphModule, graph: NNCFGraph
    ) -> Tensor:
        weight_edge = graph.get_input_edge_by_port_id(node_with_weight, weight_port_id)
        weight_node = weight_edge.from_node
        graph_weight_node = get_graph_node_by_name(model.graph, weight_node.node_name)
        weight = get_tensor_constant_from_node(graph_weight_node, model).data
        if weight is None:
            raise nncf.InternalError(f"Could not find a node in the model by name {weight_node}.")

        return Tensor(weight)

    def get_weight_dtype(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: torch.fx.GraphModule, graph: NNCFGraph
    ) -> TensorDataType:
        return self.get_weight(node_with_weight, weight_port_id, model, graph).dtype

    @staticmethod
    def get_weight_shape(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Tuple:
        weight_node = get_const_node(node_with_weight, weight_port_id, graph)
        edge = graph.get_edge(weight_node, node_with_weight)
        return tuple(edge.tensor_shape)

    def set_weight(
        self,
        node_with_weight: NNCFNode,
        weight_port_id: int,
        model: torch.fx.GraphModule,
        graph: NNCFGraph,
        weight: Tensor,
    ) -> None:
        constant_update_transformation_builder(node_with_weight, weight.data, input_port_id=weight_port_id)(model)

    def insert_adapters(
        self, wc_params: WeightCompressionParameters, lora_A: Tensor, lora_B: Tensor, int8_lora: bool
    ) -> None:
        pass

    def transform_model(
        self,
        model: torch.fx.GraphModule,
        graph: NNCFGraph,
        weight_compression_parameters: Iterable[WeightCompressionParameters],
        precomputed_scales: Dict[str, Tensor] = None,
        precomputed_zero_points: Dict[str, Tensor] = None,
        lora_correction_algo: LoraCorrectionAlgorithm = None,
    ) -> torch.fx.GraphModule:
        transformation_layout = TransformationLayout()

        for wc_params in weight_compression_parameters:
            compression_config = wc_params.compression_config
            if compression_config.mode in [
                CompressWeightsMode.NF4,
                CompressWeightsMode.E2M1,
            ]:
                raise nncf.ParameterNotSupportedError(f"{compression_config.mode.value} is not supported.")
            weight_node = get_const_node(wc_params.node_with_weight, wc_params.weight_port_id, graph)
            weight_name = weight_node.node_name
            weight = self.get_weight(wc_params.node_with_weight, wc_params.weight_port_id, model, graph)
            if weight is None or not isinstance(weight, Tensor):
                raise nncf.InternalError(f"Could not find a nncf.tensor in the model by name {weight_name}.")

            # calculates compressed weights and decompression parameters
            compressed_weight = compress_weight(
                weight,
                wc_params.reduction_axes,
                compression_config,
                None if precomputed_scales is None else precomputed_scales.get(wc_params.weight_name),
                None if precomputed_zero_points is None else precomputed_zero_points.get(wc_params.weight_name),
            )

            # creates weight decompressor
            if compression_config.mode == CompressWeightsMode.INT8_SYM:
                decompressor = INT8SymmetricWeightsDecompressor(
                    compressed_weight.scale.data, result_dtype=weight.data.dtype
                )
            elif compression_config.mode == CompressWeightsMode.INT8_ASYM:
                decompressor = INT8AsymmetricWeightsDecompressor(
                    compressed_weight.scale.data, compressed_weight.zero_point.data, result_dtype=weight.data.dtype
                )
            elif compression_config.mode == CompressWeightsMode.INT4_SYM:
                decompressor = INT4SymmetricWeightsDecompressor(
                    scale=compressed_weight.scale.data,
                    compressed_weight_shape=compressed_weight.tensor.shape,
                    result_shape=weight.shape,
                    result_dtype=weight.data.dtype,
                )
            elif compression_config.mode == CompressWeightsMode.INT4_ASYM:
                decompressor = INT4AsymmetricWeightsDecompressor(
                    scale=compressed_weight.scale.data,
                    zero_point=compressed_weight.zero_point.data,
                    compressed_weight_shape=compressed_weight.tensor.shape,
                    result_shape=weight.shape,
                    result_dtype=weight.data.dtype,
                )

            # pack tensor
            packed_tensor = decompressor.pack_weight(compressed_weight.tensor.data)

            # sets compressed tensor
            self.set_weight(wc_params.node_with_weight, wc_params.weight_port_id, model, graph, packed_tensor)

            # register weight decompression module in the model
            graph_weight_node = get_graph_node_by_name(model.graph, wc_params.node_with_weight.node_name)
            compressed_weight_name = graph_weight_node.all_input_nodes[wc_params.weight_port_id].name

            decompressor_suffix = "_".join(compressed_weight_name.replace(".", "_").split("_")[:-2])
            decompressor_name = f"{decompressor.quantization_mode}_weights_decompressor_{decompressor_suffix}"

            # inserts the weight decompressor into the model as the post hook on the model weight
            transformation_layout.register(
                FXApplyTransformationCommand(
                    module_insertion_transformation_builder(
                        decompressor,
                        [PTTargetPoint(TargetType.OPERATOR_POST_HOOK, target_node_name=compressed_weight_name)],
                        decompressor_name,
                    )
                )
            )

        # apply transformations
        transformed_model = FXModelTransformer(model).transform(transformation_layout)

        return transformed_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Iterable, List, Optional, Tuple

import openvino as ov
from openvino.runtime import opset13 as opset

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.utils import get_reduction_axes
from nncf.common.utils.caching import disable_results_caching
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import MaxVarianceTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import MeanMagnitudeTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import MeanVarianceTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.metatypes.groups import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.openvino.graph.model_transformer import OVModelTransformer
from nncf.openvino.graph.node_utils import convert_op
from nncf.openvino.graph.node_utils import create_ov_const_from_tensor
from nncf.openvino.graph.node_utils import get_const_value
from nncf.openvino.graph.node_utils import get_weight_channel_axes
from nncf.openvino.graph.transformations.command_creation import OVCommandCreator
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.optimized_functions import clear_ov_model_cache
from nncf.openvino.optimized_functions.models import OV_MODEL_CACHE
from nncf.openvino.rt_info import dump_parameters
from nncf.openvino.statistics.collectors import OVMaxVarianceReducer
from nncf.openvino.statistics.collectors import OVMeanAbsMaxReducer
from nncf.openvino.statistics.collectors import OVMeanReducer
from nncf.openvino.statistics.collectors import OVMeanVarianceReducer
from nncf.openvino.statistics.collectors import OVShapeReducer
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.awq_patterns import get_awq_patterns
from nncf.quantization.algorithms.weight_compression.backend import AWQAlgoBackend
from nncf.quantization.algorithms.weight_compression.backend import MixedPrecisionAlgoBackend
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.lora_correction import LoraCorrectionAlgorithm
from nncf.quantization.algorithms.weight_compression.weight_lowering import compress_weight
from nncf.tensor import Tensor
from nncf.tensor.definitions import TensorDataType
from nncf.tensor.functions.openvino_numeric import DTYPE_MAP_REV


class OVWeightCompressionAlgoBackend(WeightCompressionAlgoBackend):
    def __init__(self, model: ov.Model, name_to_node_mapping: Dict = None):
        if name_to_node_mapping is None:
            self.name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        else:
            self.name_to_node_mapping = name_to_node_mapping

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVMatMulMetatype]

    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.OVConvolutionMetatype,
            om.OVDepthwiseConvolutionMetatype,
            om.OVConvolutionBackpropDataMetatype,
            om.OVGroupConvolutionMetatype,
            om.OVGroupConvolutionBackpropDataMetatype,
        ]

    @property
    def embedding_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVEmbeddingMetatype]

    @staticmethod
    def is_node_with_weights(node: NNCFNode, graph: NNCFGraph) -> bool:
        return node.layer_attributes and node.layer_attributes.constant_attributes

    @staticmethod
    def get_reduction_axes(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Optional[Tuple[int]]:
        channel_axes = get_weight_channel_axes(node_with_weight)
        const_shape = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["shape"]
        return get_reduction_axes(channel_axes, const_shape)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    def mean_statistic_collector(
        self, reduction_axes: Tuple[int], subset_size: Optional[int] = None
    ) -> TensorCollector:
        mean_reducer = OVMeanReducer(reduction_axes, inplace=True)
        shape_reducer = OVShapeReducer(inplace=True)
        collector = TensorCollector(WCTensorStatistic)
        collector.register_statistic_branch(WCTensorStatistic.MEAN_STAT, mean_reducer, NoopAggregator(subset_size))
        collector.register_statistic_branch(WCTensorStatistic.SHAPE_STAT, shape_reducer, NoopAggregator(subset_size))
        return collector

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        constant_ports = node.layer_attributes.get_const_port_ids()
        activation_ports = [
            e.input_port_id for e in nncf_graph.get_input_edges(node) if e.input_port_id not in constant_ports
        ]
        assert len(activation_ports) == 1
        return activation_ports[0]

    @staticmethod
    def get_weight_names_and_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Tuple[str, int]]:
        result = []
        for weight_port_id in node.layer_attributes.get_const_port_ids():
            weight_name = node.layer_attributes.constant_attributes[weight_port_id]["name"]
            result.append((weight_name, weight_port_id))
        return result

    def get_weight(self, node_with_weight: NNCFNode, weight_port_id: int, model: ov.Model, graph: NNCFGraph) -> Tensor:
        weight_name = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["name"]
        weight_node = self.name_to_node_mapping[weight_name]
        weight_tensor = get_const_value(weight_node)
        return Tensor(weight_tensor)

    def get_weight_dtype(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: ov.Model, graph: NNCFGraph
    ) -> TensorDataType:
        ov_type_name = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["dtype"]
        ov_type = getattr(ov.Type, ov_type_name)
        return DTYPE_MAP_REV[ov_type]

    @staticmethod
    def get_weight_shape(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Tuple:
        return node_with_weight.layer_attributes.constant_attributes[weight_port_id]["shape"]

    def set_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: ov.Model, graph: NNCFGraph, weight: Tensor
    ):
        node_with_const = self.name_to_node_mapping[node_with_weight.node_name]

        const_port = node_with_const.input(weight_port_id)
        const_node = node_with_const.input_value(weight_port_id).get_node()

        shared_memory = True
        if const_node.get_element_type() == ov.Type.bf16:
            # Shared memory does not work for BF16 precision
            shared_memory = False

        new_const_node = ov.runtime.op.Constant(weight.data, shared_memory=shared_memory)
        new_const_node.set_friendly_name(const_node.get_friendly_name())
        const_port.replace_source_output(new_const_node.output(0))

        const_name = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["name"]
        self.name_to_node_mapping[const_name] = new_const_node

        new_output = new_const_node.output(0)
        for target_input in const_node.output(0).get_target_inputs():
            target_input.replace_source_output(new_output)

        del const_node

    def insert_adapters(
        self, wc_params: WeightCompressionParameters, lora_A: Tensor, lora_B: Tensor, int8_lora: bool
    ) -> None:
        input_node = self.name_to_node_mapping[wc_params.node_with_weight.node_name].input_value(0)
        activation_dtype = input_node.get_element_type()
        should_add_convert_node = activation_dtype != ov.Type.f16
        mm_node = self.name_to_node_mapping[wc_params.node_with_weight.node_name]

        if int8_lora:
            const_node_name = wc_params.node_with_weight.node_name
            int8_compression_config = WeightCompressionConfig(mode=CompressWeightsMode.INT8_ASYM, group_size=-1)
            A_W, _ = self._create_compression_subgraph(
                weight=lora_A,
                compression_config=int8_compression_config,
                reduction_axes=wc_params.reduction_axes,
                const_node_name=const_node_name + "_lora_A",
                weight_port_id=1,
                const_dtype=activation_dtype,
                should_add_convert_node=should_add_convert_node,
            )
            B_W, _ = self._create_compression_subgraph(
                weight=lora_B,
                compression_config=int8_compression_config,
                reduction_axes=wc_params.reduction_axes,
                const_node_name=const_node_name + "_lora_B",
                weight_port_id=1,
                const_dtype=activation_dtype,
                should_add_convert_node=should_add_convert_node,
            )
        else:
            A_W = opset.constant(lora_A.data)
            B_W = opset.constant(lora_B.data)

        A_MM = opset.matmul(input_node, A_W, transpose_a=False, transpose_b=True)
        B_MM = opset.matmul(A_MM, B_W, transpose_a=False, transpose_b=True)

        node_output_port = mm_node.output(0)
        node_output_source_ports = node_output_port.get_target_inputs()
        add = opset.add(mm_node, B_MM)
        for node_output_source_port in node_output_source_ports:
            node_output_source_port.replace_source_output(add.output(0))

    def _create_compression_subgraph(
        self,
        weight: Tensor,
        compression_config: WeightCompressionConfig,
        reduction_axes: Tuple[int, ...],
        const_node_name: str,
        weight_port_id: int,
        const_dtype,
        should_add_convert_node: bool,
        layer_scales: Optional[Tensor] = None,
        layer_zero_points: Optional[Tensor] = None,
    ):
        scale_dtype = ov.Type.f16
        if compression_config.mode == CompressWeightsMode.NF4:
            compression_dtype = ov.Type.nf4
        elif compression_config.mode == CompressWeightsMode.E2M1:
            compression_dtype = ov.Type.f4e2m1
            scale_dtype = ov.Type.f8e8m0
        elif compression_config.mode == CompressWeightsMode.INT4_SYM:
            compression_dtype = ov.Type.i4
        elif compression_config.mode == CompressWeightsMode.INT4_ASYM:
            compression_dtype = ov.Type.u4
        elif compression_config.mode == CompressWeightsMode.INT8_SYM:
            compression_dtype = ov.Type.i8
        elif compression_config.mode == CompressWeightsMode.INT8_ASYM:
            compression_dtype = ov.Type.u8
        else:
            raise nncf.ParameterNotSupportedError(f"{compression_config.mode.value} is not supported.")

        original_shape = weight.shape
        with disable_results_caching(OV_MODEL_CACHE):
            compressed_weight = compress_weight(
                weight,
                reduction_axes,
                compression_config,
                layer_scales,
                layer_zero_points,
            )

        compressed_const = create_ov_const_from_tensor(
            compressed_weight.tensor, compression_dtype, name=const_node_name
        )
        converted_const = opset.convert(compressed_const, ov.Type.f16)

        if compressed_weight.zero_point is not None:
            zero_point_const = create_ov_const_from_tensor(
                compressed_weight.zero_point, compression_dtype, name=f"{const_node_name}/zero_point"
            )
            zero_point_const = opset.convert(zero_point_const, ov.Type.f16)
            converted_const = opset.subtract(
                converted_const, zero_point_const, name=f"{const_node_name}/zero_point/subtract"
            )

        scale_const = create_ov_const_from_tensor(compressed_weight.scale, scale_dtype, name=f"{const_node_name}/scale")
        scale_const = convert_op(scale_const, ov.Type.f16)

        mul = opset.multiply(
            converted_const,
            scale_const,
            name=f"{const_node_name}/fq_weights_{weight_port_id}",
        )

        if compression_config.group_size != -1:
            mul = opset.reshape(mul, output_shape=original_shape, special_zero=False)

        if should_add_convert_node:
            mul = opset.convert(mul, const_dtype, name=f"{const_node_name}/fq_weights_{weight_port_id}/convert")
        return mul, compressed_weight

    def transform_model(
        self,
        model: ov.Model,
        graph: NNCFGraph,
        weight_compression_parameters: Iterable[WeightCompressionParameters],
        precomputed_scales: Dict[str, Tensor] = None,
        precomputed_zero_points: Dict[str, Tensor] = None,
        lora_correction_algo: LoraCorrectionAlgorithm = None,
    ) -> ov.Model:
        for wc_params in weight_compression_parameters:
            const_attributes = wc_params.node_with_weight.layer_attributes.constant_attributes[wc_params.weight_port_id]
            const_node_name = const_attributes["name"]
            const_node = self.name_to_node_mapping[const_node_name]
            const_node_output = const_node.output(0)
            const_dtype = const_node_output.get_element_type()
            weight = get_const_value(const_node, cast_bf16_to_fp32=False)
            # Creation of ov.Tensor is required for two reasons:
            #   1. To be able to process BF16 weight properly
            #   2. To indicate that it is allowed for the compressed constant to be returned as int4/uint4 if needed
            weight = ov.Tensor(weight, weight.shape, const_dtype)
            weight = Tensor(weight)

            should_add_convert_node = False
            if const_dtype != ov.Type.f16:
                for inp in const_node_output.get_target_inputs():
                    if inp.get_node().get_type_name() != "Convert":
                        should_add_convert_node = True
                        break

            layer_scales = None if precomputed_scales is None else precomputed_scales.get(wc_params.weight_name)
            layer_zero_points = (
                None if precomputed_zero_points is None else precomputed_zero_points.get(wc_params.weight_name)
            )
            mul, compressed_weight = self._create_compression_subgraph(
                weight=weight,
                compression_config=wc_params.compression_config,
                reduction_axes=wc_params.reduction_axes,
                const_node_name=const_node_name,
                weight_port_id=wc_params.weight_port_id,
                const_dtype=const_dtype,
                should_add_convert_node=should_add_convert_node,
                layer_scales=layer_scales,
                layer_zero_points=layer_zero_points,
            )

            mul_output = mul.output(0)
            for target_input in const_node.output(0).get_target_inputs():
                target_input.replace_source_output(mul_output)
            if lora_correction_algo is not None and lora_correction_algo.is_applicable(wc_params):
                # These tensors can potentially be in ov backend
                weight = weight.as_numpy_tensor()
                compressed_weight.tensor = compressed_weight.tensor.as_numpy_tensor()
                if compressed_weight.zero_point is not None:
                    compressed_weight.zero_point = compressed_weight.zero_point.as_numpy_tensor()
                adapters = lora_correction_algo.calculate_adapters(weight, compressed_weight, wc_params)
                self.insert_adapters(wc_params, *adapters, int8_lora=lora_correction_algo.use_int8_adapters)

        # reset name_to_node_mapping
        self.name_to_node_mapping = None

        clear_ov_model_cache()

        return model

    @staticmethod
    def dump_parameters(
        model: ov.Model, parameters: Dict, algo_name: Optional[str] = "quantization", path: Optional[List] = None
    ) -> None:
        dump_parameters(model, parameters, algo_name, path)


class OVAWQAlgoAlgoBackend(AWQAlgoBackend, OVWeightCompressionAlgoBackend):
    @staticmethod
    def get_awq_patterns():
        return get_awq_patterns(om.OVMatMulMetatype, om.OVMultiplyMetatype, ATOMIC_ACTIVATIONS_OPERATIONS)

    @staticmethod
    def scale_insertion_command(source_node, next_nodes, source_node_output_port, scale):
        return OVCommandCreator.multiply_insertion_command(
            source_node, next_nodes, source_node_output_port, scale, f"{source_node.node_name}/awq_mul"
        )


class OVMixedPrecisionAlgoBackend(MixedPrecisionAlgoBackend, OVWeightCompressionAlgoBackend):
    @staticmethod
    def mean_variance_statistic_collector(
        reduction_axes: Tuple[int], subset_size: Optional[int] = None
    ) -> TensorCollector:
        reducer = OVMeanVarianceReducer(reduction_axes, inplace=True)
        aggregator = MeanAggregator(num_samples=subset_size)
        collector = TensorCollector(MeanVarianceTensorStatistic)
        collector.register_statistic_branch(MeanVarianceTensorStatistic.MEAN_VARIANCE_STAT, reducer, aggregator)
        return collector

    @staticmethod
    def max_variance_statistic_collector(
        reduction_axes: Tuple[int], subset_size: Optional[int] = None
    ) -> TensorCollector:
        reducer = OVMaxVarianceReducer(reduction_axes, inplace=True)
        aggregator = MeanAggregator(num_samples=subset_size)
        collector = TensorCollector(MaxVarianceTensorStatistic)
        collector.register_statistic_branch(MaxVarianceTensorStatistic.MAX_VARIANCE_STAT, reducer, aggregator)
        return collector

    @staticmethod
    def mean_abs_max_statistic_collector(
        reduction_axes: Tuple[int], subset_size: Optional[int] = None
    ) -> TensorCollector:
        reducer = OVMeanAbsMaxReducer(reduction_axes, inplace=True)
        aggregator = MeanAggregator(num_samples=subset_size)
        collector = TensorCollector(MeanMagnitudeTensorStatistic)
        collector.register_statistic_branch(MeanMagnitudeTensorStatistic.MEAN_MAGNITUDE_STAT, reducer, aggregator)
        return collector



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from collections import deque
from typing import Callable, Dict, List, Tuple

import numpy as np
import openvino.runtime as ov
from openvino._pyopenvino import DescriptorTensor
from openvino.runtime import opset13 as opset

import nncf
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.model_transformer import TModel
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.openvino.graph.model_utils import copy_rt_info
from nncf.openvino.graph.node_utils import get_parameter_node_name
from nncf.openvino.graph.node_utils import get_result_node_name
from nncf.openvino.graph.transformations.commands import OVBiasCorrectionCommand
from nncf.openvino.graph.transformations.commands import OVBiasInsertionCommand
from nncf.openvino.graph.transformations.commands import OVConvertInsertionCommand
from nncf.openvino.graph.transformations.commands import OVExtractIfBodyCommand
from nncf.openvino.graph.transformations.commands import OVFQNodeRemovingCommand
from nncf.openvino.graph.transformations.commands import OVInplaceFnInsertionCommand
from nncf.openvino.graph.transformations.commands import OVModelExtractionCommand
from nncf.openvino.graph.transformations.commands import OVMultiplyInsertionCommand
from nncf.openvino.graph.transformations.commands import OVOutputInsertionCommand
from nncf.openvino.graph.transformations.commands import OVQuantizerInsertionCommand
from nncf.openvino.graph.transformations.commands import OVStateLessModelExtractionCommand
from nncf.openvino.graph.transformations.commands import OVUpdateIfBodyCommand
from nncf.openvino.graph.transformations.commands import OVWeightUpdateCommand
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters


class OVModelTransformer(ModelTransformer):
    """
    Applies transformations to an OpenVINO model.
    """

    def __init__(self, model: TModel, inplace: bool = False):
        super().__init__(model)
        self._inplace = inplace
        self._command_transformation_ordered_pairs = [
            (OVFQNodeRemovingCommand, self._apply_fq_nodes_removing_transformation),
            (OVQuantizerInsertionCommand, self._apply_quantizer_insertion_transformations),
            (OVConvertInsertionCommand, self._apply_convert_insertion_transformations),
            (OVBiasCorrectionCommand, self._apply_bias_correction_transformations),
            (OVWeightUpdateCommand, self._apply_weight_update_transformations),
            (OVModelExtractionCommand, self._apply_model_extraction_transformation),
            (OVStateLessModelExtractionCommand, self._apply_stateless_model_extraction_transformation),
            (OVInplaceFnInsertionCommand, self._apply_insert_operation),
            (OVOutputInsertionCommand, self._apply_output_insertion_transformations),
            (OVBiasInsertionCommand, self._apply_bias_insertion_transformations),
            (OVMultiplyInsertionCommand, self._apply_multiply_insertion_transformations),
            (OVUpdateIfBodyCommand, self._apply_update_if_body_transformations),
            (OVExtractIfBodyCommand, self._apply_extract_if_body_transformation),
        ]

    @staticmethod
    def _convert_to_dtype(value, dtype):
        clip_data = np.clip(value, np.finfo(dtype).min, np.finfo(dtype).max)
        return clip_data.astype(dtype)

    @staticmethod
    def _get_name_to_node_mapping(model: ov.Model) -> Dict[str, ov.Node]:
        """
        Returns name to node mapping.

        :param model: Model to get mapping.
        :return: Mapping from node name to node.
        """
        return {op.get_friendly_name(): op for op in model.get_ops()}

    @staticmethod
    def _get_activation_node_names(model: ov.Model) -> List[str]:
        """
        Returns list of the activation node names.

        :param model: Model to get list.
        :return: List with the activation names.
        """
        activation_nodes = set()
        nodes_queue = deque(model.get_parameters())
        while nodes_queue:
            node = nodes_queue.popleft()
            if node.name in activation_nodes:
                continue
            activation_nodes.add(node.name)
            for node_output in node.outputs():
                nodes_queue.extend([i.get_node() for i in node_output.get_target_inputs()])
        return list(activation_nodes)

    @staticmethod
    def _update_tensor_names(tensors: List[DescriptorTensor], names: List[str]) -> None:
        """
        Updates tensors names in-place.

        :param model: List of the tensors.
        :param names: List of the new names for tensors.
        """
        for tensor in tensors:
            current_names = tensor.get_names()
            current_names.update(names)
            tensor.set_names(current_names)

    def transform(self, transformation_layout: TransformationLayout) -> ov.Model:
        """
        Applies transformations to the model using an out-of-place approach.
        The transformations do not affect the original model, and a new model
        is returned with the transformations applied. If there are no transformations,
        returns a new instance of the original model.

        :param transformation_layout: Transformation commands.
        :return: The new instance of a model with applied transformations.
        """

        transformations = transformation_layout.transformations
        aggregated_transformations = defaultdict(list)
        for transformation in transformations:
            aggregated_transformations[transformation.__class__].append(transformation)

        if self._inplace:
            model = self._model
        else:
            model = self._model.clone()
        # Inplace transformations; Using deepcopy of model
        for transformation_cls, transformation_fn in self._command_transformation_ordered_pairs:
            transformations = aggregated_transformations[transformation_cls]
            if transformations:
                model = transformation_fn(model, transformations)

        return model

    @staticmethod
    def _apply_output_insertion_transformations(
        model: ov.Model, transformations: List[OVOutputInsertionCommand]
    ) -> ov.Model:
        """
        Applies incoming transformations to the model.

        :param model: Model to apply transformations.
        :param transformations: OVOutputInsertionCommand transformations.
        :return: Model with inserted outputs.
        """
        extra_model_outputs = OVModelTransformer._get_extra_model_outputs(model, transformations)
        return OVModelTransformer._insert_outputs(model, outputs=extra_model_outputs)

    @staticmethod
    def _get_extra_model_outputs(
        model: ov.Model, transformations: List[OVOutputInsertionCommand]
    ) -> List[Tuple[ov.Output, int]]:
        """
        Collects extra model outputs based on transformations.

        :param transformations: lisf of the OVOutputInsertionCommand.
        :return: list of tuples with ov.Output & port_id.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        extra_model_outputs = []
        for transformation in transformations:
            node_name = transformation.target_point.target_node_name
            node = name_to_node_mapping[node_name]
            port_id = transformation.target_point.port_id
            output_dtype = transformation.output_dtype
            if transformation.target_point.type == TargetType.POST_LAYER_OPERATION:
                output = node.output(port_id)
                extra_model_outputs.append((output, port_id, output_dtype))
            elif transformation.target_point.type in [
                TargetType.PRE_LAYER_OPERATION,
                TargetType.OPERATION_WITH_WEIGHTS,
            ]:
                output = node.input_value(port_id)
                extra_model_outputs.append((output, output.get_index(), output_dtype))
            else:
                raise NotImplementedError(f"Unsupported target point type {transformation.target_point.type}")

        return extra_model_outputs

    @staticmethod
    def _insert_outputs(model: ov.Model, outputs: List[Tuple[ov.Output, int, Callable[[str, int], str]]]) -> ov.Model:
        """
        Takes a model and adds outputs based on the list of ov.Output.

        :param model: OpenVINO model.
        :param outputs: list of tuples with ov.Output & port_id.
        :return: Model with new outputs.
        """
        results = model.get_results()
        params = model.get_parameters()

        extra_model_outputs = []
        for output, port_id, dtype in outputs:
            node_output = output
            output_name = node_output.get_node().get_friendly_name()
            result_name = get_result_node_name(output_name, port_id)

            if dtype is not None and node_output.get_element_type() != dtype:
                node_output = opset.convert(output, destination_type=dtype)

            result = opset.result(node_output, name=result_name)
            result_tensor_names = [result_name] + list(output.get_names())
            OVModelTransformer._update_tensor_names([result.get_output_tensor(0)], result_tensor_names)
            extra_model_outputs.append(result)

        model_with_outputs = ov.Model(
            results=results + extra_model_outputs, sinks=model.get_sinks(), parameters=params, name=model.friendly_name
        )
        copy_rt_info(model, model_with_outputs, path=["nncf"])
        return model_with_outputs

    @staticmethod
    def _apply_fq_nodes_removing_transformation(
        model: ov.Model, transformations: List[OVFQNodeRemovingCommand]
    ) -> ov.Model:
        """
        Removes the layers from the model.

        :param model: Model to apply transformations.
        :param transformations: Node removing transformations.
        :return: Model with removed FakeQuantize nodes.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            node = name_to_node_mapping[transformation.target_point.target_node_name]

            node_input = node.input_value(0)
            for node_output in node.outputs():
                for target_in in node_output.get_target_inputs():
                    target_in.replace_source_output(node_input)
            del name_to_node_mapping[transformation.target_point.target_node_name]
        return model

    @staticmethod
    def _apply_quantizer_insertion_transformations(
        model: ov.Model, transformations: List[OVQuantizerInsertionCommand]
    ) -> ov.Model:
        """
        Applies transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: List of the OVQuantizerInsertionCommand transformations.
        :return: Model with inserted FakeQuantize nodes.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            OVModelTransformer._insert_fake_quantize_op(transformation, name_to_node_mapping)
        return model

    @staticmethod
    def _apply_convert_insertion_transformations(
        model: ov.Model, transformations: List[OVConvertInsertionCommand]
    ) -> ov.Model:
        """
        Applies transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: List of the OVConvertInsertionCommand transformations.
        :return: Model with inserted FakeConvert nodes.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            OVModelTransformer._insert_fake_convert_op(transformation, name_to_node_mapping)
        return model

    @staticmethod
    def _create_constant(value: np.ndarray, dtype: ov.Type, name: str, shared_memory: bool = False) -> ov.Node:
        """
        Creates constant using opset.

        :param value: Numpy value.
        :param type: Constant type.
        :param name: Name for the constant.
        :param shared_memory: Shared memory option.
        :return: ov.Node instance.
        """
        constant_value = value
        # BF16 does not support type conversion in numpy
        if dtype != ov.Type.bf16 and constant_value.dtype != dtype.to_dtype():
            constant_value = OVModelTransformer._convert_to_dtype(constant_value, dtype.to_dtype())

        return opset.constant(constant_value, dtype=dtype, name=name, shared_memory=shared_memory)

    @staticmethod
    def _create_fake_quantize(
        op_output: ov.Output,
        fake_quantize_params: FakeQuantizeParameters,
        fake_quantize_name: str,
        data_type: ov.Type,
    ) -> ov.Node:
        """
        Creates FakeQuantize node.

        :param op_output: Output of the previous node.
        :param fake_quantize_params: FakeQuantizeParameters instance.
        :param fake_quantize_name: New layer name.
        :param data_type: ov.Type instance for data.
        :return: ov.Node instance.
        """

        input_low = fake_quantize_params.input_low.data
        input_high = fake_quantize_params.input_high.data
        output_low = fake_quantize_params.output_low.data
        output_high = fake_quantize_params.output_high.data
        levels = fake_quantize_params.levels

        input_low = OVModelTransformer._create_constant(
            input_low, dtype=data_type, name=f"{fake_quantize_name}/input_low"
        )
        input_high = OVModelTransformer._create_constant(
            input_high, dtype=data_type, name=f"{fake_quantize_name}/input_high"
        )
        output_low = OVModelTransformer._create_constant(
            output_low, dtype=data_type, name=f"{fake_quantize_name}/output_low"
        )
        output_high = OVModelTransformer._create_constant(
            output_high, dtype=data_type, name=f"{fake_quantize_name}/output_high"
        )

        return opset.fake_quantize(
            op_output, input_low, input_high, output_low, output_high, levels, name=fake_quantize_name
        )

    @staticmethod
    def _create_fake_convert(
        op_output: ov.Output,
        fake_convert_params: FakeConvertParameters,
        fake_convert_name: str,
        data_type: ov.Type,
    ) -> ov.Node:
        """
        Creates FakeConvert node.

        :param op_output: Output of the previous node.
        :param fake_convert_params: FakeConvertParameters instance.
        :param fake_convert_name: New layer name.
        :param data_type: ov.Type instance for data.
        :return: ov.Node instance.
        """

        scale = fake_convert_params.scale.data
        shift = fake_convert_params.shift.data

        destination_type = fake_convert_params.destination_type.value
        scale = OVModelTransformer._create_constant(scale, dtype=data_type, name=f"{fake_convert_name}/scale")
        shift = OVModelTransformer._create_constant(shift, dtype=data_type, name=f"{fake_convert_name}/shift")

        return opset.fake_convert(
            data=op_output,
            scale=scale,
            shift=shift,
            destination_type=destination_type,
            name=fake_convert_name,
        )

    @staticmethod
    def _insert_fake_quantize_op(
        transformation: OVQuantizerInsertionCommand, name_to_node_mapping: Dict[str, ov.Node]
    ) -> None:
        """
        Inserts FakeQuantize Operation to a model which name_to_node_mapping is passed.

        :param transformation: FakeQuantize insertion command.
        :param name_to_node_mapping: Mapping from node name to node instance.
        """
        fq_params = transformation.quantizer_parameters

        node_name = transformation.target_point.target_node_name
        target_node = name_to_node_mapping[node_name]
        port_id = transformation.target_point.port_id
        transform_type = transformation.target_point.type
        if transform_type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
            inp_node = target_node.input(port_id)
            input_node_output = inp_node.get_source_output()
            data_type = inp_node.get_element_type()
            name = "fq_weights" if transform_type == TargetType.OPERATION_WITH_WEIGHTS else "fq_input"
            fq_name = f"{node_name}/{name}_{port_id}"

            fq = None
            if transform_type == TargetType.OPERATION_WITH_WEIGHTS:
                # If the nodes share one weight tensor, we should have only one quantizer on that
                for out in input_node_output.get_target_inputs():
                    if out.get_node().get_type_name() == "FakeQuantize":
                        fq = out.get_node()
            if fq is None:
                fq = OVModelTransformer._create_fake_quantize(
                    op_output=input_node_output,
                    fake_quantize_params=fq_params,
                    fake_quantize_name=fq_name,
                    data_type=data_type,
                )
            inp_node.replace_source_output(fq.output(0))
        elif transform_type == TargetType.POST_LAYER_OPERATION:
            output = target_node.output(port_id)
            data_type = output.get_element_type()
            target_inputs = output.get_target_inputs()
            fq_name = f"{node_name}/fq_output_{port_id}"
            fq = OVModelTransformer._create_fake_quantize(
                op_output=output,
                fake_quantize_params=fq_params,
                fake_quantize_name=fq_name,
                data_type=data_type,
            )
            for inp_node in target_inputs:
                inp_node.replace_source_output(fq.output(0))
        else:
            raise nncf.InternalError(f"Incorrect target point type {transform_type}")

    @staticmethod
    def _insert_fake_convert_op(
        transformation: OVConvertInsertionCommand, name_to_node_mapping: Dict[str, ov.Node]
    ) -> None:
        """
        Inserts FakeConvert Operation to a model which name_to_node_mapping is passed.

        :param transformation: FakeConvert insertion command.
        :param name_to_node_mapping: Mapping from node name to node instance.
        """
        fc_params = transformation.convert_parameters

        node_name = transformation.target_point.target_node_name
        target_node = name_to_node_mapping[node_name]
        port_id = transformation.target_point.port_id
        transform_type = transformation.target_point.type
        name = "weights" if transform_type == TargetType.OPERATION_WITH_WEIGHTS else "input"

        if transform_type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
            inp_node = target_node.input(port_id)
            input_node_output = inp_node.get_source_output()

            fc = None
            if transform_type == TargetType.OPERATION_WITH_WEIGHTS:
                # If the nodes share one weight tensor, we should have only one quantizer on that
                for out in input_node_output.get_target_inputs():
                    if out.get_node().get_type_name() == "FakeConvert":
                        fc = out.get_node()
            if fc is None:
                data_type = inp_node.get_element_type()
                fc_name = f"{node_name}/fc_{name}_{port_id}"
                fc = OVModelTransformer._create_fake_convert(
                    op_output=input_node_output,
                    fake_convert_params=fc_params,
                    fake_convert_name=fc_name,
                    data_type=data_type,
                )
            inp_node.replace_source_output(fc.output(0))
        elif transform_type == TargetType.POST_LAYER_OPERATION:
            output = target_node.output(port_id)
            data_type = output.get_element_type()
            target_inputs = output.get_target_inputs()
            fc_name = f"{node_name}/fc_output_{port_id}"
            fc = OVModelTransformer._create_fake_convert(
                op_output=output,
                fake_convert_params=fc_params,
                fake_convert_name=fc_name,
                data_type=data_type,
            )
            for inp_node in target_inputs:
                inp_node.replace_source_output(fc.output(0))
        else:
            raise nncf.InternalError(f"Incorrect target point type {transform_type}")

    @staticmethod
    def _apply_bias_correction_transformations(model, transformations: List[OVBiasCorrectionCommand]) -> ov.Model:
        """
        Applies bias correction transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: List of the bias correction transformations.
        :return: Model with corrected bias.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            node = name_to_node_mapping[transformation.target_point.target_node_name]
            node_inputs = [port.get_node() for port in node.output(0).get_target_inputs()]
            assert any(node.get_type_name() == "Add" for node in node_inputs)

            for node_input in node_inputs:
                if node_input.get_type_name() == "Add":
                    add_node = node_input

            OVModelTransformer._set_const_value(
                add_node, transformation.target_point.port_id, transformation.bias_value
            )
        return model

    @staticmethod
    def _set_const_value(node_with_const: ov.Node, const_port_id: int, const_value: np.ndarray) -> None:
        port = node_with_const.input(const_port_id)
        node = node_with_const.input_value(const_port_id).get_node()

        const_port = None
        const_node = None
        queue = deque([(port, node)])
        while len(queue) != 0:
            curr_port, curr_node = queue.popleft()
            if curr_node.get_type_name() == "Constant":
                const_port = curr_port
                const_node = curr_node
                break
            if len(curr_node.inputs()) == 0:
                break
            queue.append((curr_node.input(0), curr_node.input_value(0).get_node()))

        if const_node is None:
            raise nncf.InternalError("Constant node was expected but could not find it.")

        const_value = np.reshape(const_value, const_node.data.shape)

        shared_memory = True
        if const_port.get_element_type() == ov.Type.bf16:
            # Shared memory does not work for BF16 precision
            shared_memory = False

        new_const_node = OVModelTransformer._create_constant(
            const_value,
            dtype=const_port.get_element_type(),
            name=const_node.get_friendly_name(),
            shared_memory=shared_memory,
        )
        const_port.replace_source_output(new_const_node.output(0))

    @staticmethod
    def _apply_weight_update_transformations(model, transformations: List[OVWeightUpdateCommand]) -> ov.Model:
        """
        Applies weight update transformation to the model.

        :param transformations: List of the weight update transformations.
        :returns: Transformed model.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            node_with_weight = name_to_node_mapping[transformation.target_point.target_node_name]
            OVModelTransformer._set_const_value(
                node_with_weight, transformation.target_point.port_id, transformation.weight_value  # Weight port id
            )
        return model

    @staticmethod
    def _apply_model_extraction_transformation(
        model: ov.Model, transformations: List[OVModelExtractionCommand]
    ) -> ov.Model:
        """
        Extracts sub-model from the original based on the inputs and outputs names.

        :param model: Model to apply transformations.
        :param transformation: Model extraction transformation.
        :return: Extracted sub-model.
        """
        outputs_type = ov.Type.f32
        transformation = transformations[-1]
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)

        params, results = [], []
        for input_name, input_port_id in transformation.input_ids:
            input_node = name_to_node_mapping[input_name]
            if input_name in [tensor.node.get_friendly_name() for tensor in model.inputs]:
                params.append(input_node)
                continue

            input_port = input_node.input(input_port_id)
            input_type = input_port.get_element_type()
            input_node_output = input_port.get_source_output()
            parameter_name = get_parameter_node_name(input_name, input_port_id)

            new_param = opset.parameter(
                shape=input_node_output.partial_shape,
                dtype=outputs_type,
                name=parameter_name,
            )
            new_input = new_param.output(0)

            if input_type != outputs_type:
                new_input = opset.convert(new_param, destination_type=input_type).output(0)

            input_port.replace_source_output(new_input)
            new_param_tensors = [o.get_tensor() for o in new_param.outputs()]
            OVModelTransformer._update_tensor_names(new_param_tensors, [parameter_name])
            params.append(new_param)

        for output_name, output_port_id in transformation.output_ids:
            output_node = name_to_node_mapping[output_name]

            result_name = get_result_node_name(output_name, output_port_id)
            output_port = output_node.output(output_port_id)
            if output_port.get_element_type() != outputs_type:
                output_port = opset.convert(output_node, destination_type=outputs_type).output(0)
            new_result = opset.result(output_port, name=result_name)
            result_tensor_names = [result_name] + list(output_port.get_names())
            OVModelTransformer._update_tensor_names([new_result.get_output_tensor(0)], result_tensor_names)
            results.append(new_result)

        if not results:
            results = model.get_results()

        extracted_model = ov.Model(results, params)
        copy_rt_info(model, extracted_model, path=["nncf"])
        return extracted_model

    @staticmethod
    def _apply_stateless_model_extraction_transformation(
        model: ov.Model, transformations: List[OVStateLessModelExtractionCommand]
    ) -> ov.Model:
        """
        Extracts stateless sub-model from the original based on the inputs and outputs names.

        :param model: Model to apply transformations.
        :param transformation: Model extraction transformation.
        :return: Extracted sub-model.
        """
        transformation = transformations[-1]
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)

        params, results = [], []
        model_input_names = [tensor.node.get_friendly_name() for tensor in model.inputs]
        for input_name, output_port_id in transformation.input_ids:
            input_node = name_to_node_mapping[input_name]
            if input_name in model_input_names:
                params.append(input_node)
                continue

            output_port = input_node.output(output_port_id)
            parameter_name = get_parameter_node_name(input_name, output_port_id)
            new_param = opset.parameter(
                shape=output_port.partial_shape,
                dtype=output_port.get_element_type(),
                name=parameter_name,
            )
            for input_port in output_port.get_target_inputs():
                input_port.replace_source_output(new_param.output(0))
            new_param_tensors = [o.get_tensor() for o in new_param.outputs()]
            OVModelTransformer._update_tensor_names(new_param_tensors, [parameter_name])
            params.append(new_param)

        for output_name, output_port_id in transformation.output_ids:
            output_node = name_to_node_mapping[output_name]

            output_port = output_node.output(output_port_id)
            result_name = get_result_node_name(output_name, output_port_id)
            new_result = opset.result(output_port, name=result_name)
            OVModelTransformer._update_tensor_names([new_result.get_output_tensor(0)], [result_name])
            results.append(new_result)

        if not results:
            results = model.get_results()

        extracted_model = ov.Model(results=results, parameters=params)
        copy_rt_info(model, extracted_model, path=["nncf"])

        for op in extracted_model.get_ops():
            if op.get_type_name() != "ReadValue":
                continue

            op_input_values = op.input_values()
            op_outputs = op.outputs()
            if op_input_values:
                for op_output in op_outputs:
                    for target_in in op_output.get_target_inputs():
                        target_in.replace_source_output(op_input_values[0])
            else:
                raise RuntimeError("ReadValue has no initial value.")

        return extracted_model

    @staticmethod
    def _apply_insert_operation(model: ov.Model, transformations: OVInplaceFnInsertionCommand) -> ov.Model:
        """
        Applies inplace fn insertion transformation to the model.

        :param transformations: lisf of the OVInplaceFnInsertionCommand.
        :returns: Transformed model.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        outputs = []
        for transformation in transformations:
            outputs.append(OVModelTransformer._insert_inplace_operation(transformation, name_to_node_mapping))
        return OVModelTransformer._insert_outputs(model, outputs)

    @staticmethod
    def _insert_inplace_operation(
        transformation: OVInplaceFnInsertionCommand, name_to_node_mapping: Dict[str, ov.Node]
    ) -> Tuple[ov.Output, int]:
        """
        Inserts operation inplace to a model which name_to_node_mapping is passed.

        :param transformation: Inplace fn insertion command.
        :param name_to_node_mapping: Mapping from node name to node instance.
        :returns: Pair with inserted node output and corresponded output port id.
        """
        transform_type = transformation.target_point.type

        node_name = transformation.target_point.target_node_name
        target_node = name_to_node_mapping[node_name]
        port_id = transformation.target_point.port_id
        fn_output_port_id = transformation.fn_output_port_id
        output_dtype = transformation.output_dtype
        if transform_type == TargetType.POST_LAYER_OPERATION:
            new_node = transformation.inplace_op_fn(target_node, port_id, transformation.last_inplace_node_name)
            return (new_node.output(fn_output_port_id), fn_output_port_id, output_dtype)
        if transform_type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
            output = target_node.input_value(port_id)
            new_node = transformation.inplace_op_fn(
                output.get_node(), output.get_index(), transformation.last_inplace_node_name
            )
            return (new_node.output(fn_output_port_id), fn_output_port_id, output_dtype)
        raise nncf.InternalError(f"Transform type {transform_type} is not supported")

    @staticmethod
    def _apply_bias_insertion_transformations(
        model: ov.Model, transformations: List[OVBiasInsertionCommand]
    ) -> ov.Model:
        """
        Inserts bias operation after corresponding layer.

        :param transformations: List of the bias insertion transformations.
        :returns: Transformed model with null biases.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            node_name = transformation.target_point.target_node_name
            node = name_to_node_mapping[node_name]
            # Since layers that may have biases mostly are Convolution or MatMul variations,
            # we may use only 0 output port.
            node_output_port = node.output(transformation.target_point.port_id)
            node_output_source_ports = node_output_port.get_target_inputs()

            bias_node_name = f"{node_name}/nncf_null_bias_"

            bias_const_node = OVModelTransformer._create_constant(
                transformation.bias_value,
                dtype=node_output_port.get_element_type(),
                name=f"{bias_node_name}/bias_value",
            )
            bias_const_output_port = bias_const_node.output(0)

            add_node = opset.add(node_output_port, bias_const_output_port, name=bias_node_name)

            for node_output_source_port in node_output_source_ports:
                node_output_source_port.replace_source_output(add_node.output(0))

        return model

    @staticmethod
    def _apply_multiply_insertion_transformations(
        model: ov.Model, transformations: List[OVMultiplyInsertionCommand]
    ) -> ov.Model:
        """
        Inserts Multiply with provided value for corresponding layer.

        :param transformations: List of the smooth insertion transformations.
        :returns: Transformed model with Multiply nodes.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)

        for transformation in transformations:
            node_name = transformation.target_point.target_node_name
            node = name_to_node_mapping[node_name]
            output_port_id = transformation.target_point.port_id
            node_output_port = node.output(output_port_id)

            destination_ports = []

            for target_input_port in node_output_port.get_target_inputs():
                target_node = target_input_port.get_node()
                if target_node.get_friendly_name() in transformation.destination_node_names:
                    destination_ports.append(target_input_port)

            scale_dtype = node_output_port.get_element_type()

            scale_constant = OVModelTransformer._create_constant(
                transformation.scale_value, dtype=scale_dtype, name=f"{transformation.multiply_node_name}/scale"
            )
            multiply_node = opset.multiply(node_output_port, scale_constant, name=transformation.multiply_node_name)

            for destination_port in destination_ports:
                destination_port.replace_source_output(multiply_node.output(0))

        return model

    @staticmethod
    def _apply_update_if_body_transformations(
        model: ov.Model, transformations: List[OVUpdateIfBodyCommand]
    ) -> ov.Model:
        """
        Update model body for IF node.

        :param model: Model to update and insert a new subgraph.
        :param transformations: Transformations with information of If node and an updated subgraph.
        :return: Original model with an updated subgraph.
        """
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        for transformation in transformations:
            subgraph_model = transformation.subgraph_model
            port_id = transformation.target_point.port_id
            node_name = transformation.target_point.target_node_name
            node = name_to_node_mapping[node_name]
            node.set_function(port_id, subgraph_model)
        return model

    @staticmethod
    def _apply_extract_if_body_transformation(
        model: ov.Model, transformations: List[OVExtractIfBodyCommand]
    ) -> ov.Model:
        """
        Extract a model body from If node.

        :param model: Model from which extracts a subgraph.
        :param transformations: Transformations with information from which
        If node and input port extract a model subgraph.
        :return: Model subgraph.
        """
        transformation = transformations[-1]
        name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        ov_node = name_to_node_mapping[transformation.if_node_name]
        if transformation.if_body_condition:
            return ov_node.get_function(0)
        return ov_node.get_function(1)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.activation_stats import process_stats
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import calculate_normalized_weight_and_fp4_scale
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_int_quantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_nf4_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_nf4_quantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import quantize_dequantize_weight
from nncf.quantization.algorithms.weight_compression.weight_lowering import reshape_weight_for_grouped_quantization
from nncf.tensor import Tensor
from nncf.tensor import TensorDataType
from nncf.tensor import functions as fns

TModel = TypeVar("TModel")


class ScaleEstimation:
    """
    Scale estimation algorithm implementation.
    """

    def __init__(
        self,
        model: TModel,
        name_to_node_mapping: Dict[str, Any],
        all_weight_params: List[WeightCompressionParameters],
        nodes_to_compress: List[NNCFNode],
        statistics: Dict[str, WCTensorStatistic],
        subset_size: int = 32,
        initial_steps: int = 5,
        scale_steps: int = 10,
        weight_penalty: float = -1.0,
    ):
        """
        :param model: Model for applying algorithm.
        :param name_to_node_mapping: Name to node mapping for updating node weights.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param statistics: Input activation statistics for each node.
        :param subset_size: The number of samples for scale estimation.
        :param initial_steps: The number of the steps for absmax scale rectification.
        :param scale_steps: The number of the steps for grid search scale rectification
                            from 1.0 to 1.0 - 0.05 * scale_step.
        :param weight_penalty: coefficient for penalty between fp and compressed weights. If -1 then doesn't apply.
        """
        super().__init__()
        self.name_to_node_mapping = name_to_node_mapping
        self._all_weight_params = all_weight_params
        self._nodes_to_compress = nodes_to_compress
        self._statistics = statistics
        self._subset_size = subset_size
        self._initial_steps = initial_steps
        self._scale_steps = scale_steps
        self._weight_penalty = weight_penalty

        self._set_backend_entity(model)

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param activations: The input activations of the layers considered for compression.
        """

        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model, self.name_to_node_mapping)
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific AWQ entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> Tuple[Dict[str, Tensor], Dict[str, Tensor]]:
        """
        Estimates better scale for the int4 nodes in the model.
        Minimizes per-group difference between floating point MatMul and
        MatMul with compressed weights.
        The algorithm computes weighted scale for the group of weights in MatMul, which
        shared the same scale.

        :param model: Model for applying algorithm.
        :param graph: Model graph.
        :param statistic_points: Statistic points with collected statistics values.
        :param dataset: A representative dataset for the calibration process.
        :return: Two dictionaries for estimated scales and zero points for each weight name.
        """

        scales, zero_points = dict(), dict()

        for wp in track(self._all_weight_params, description="Applying Scale Estimation"):
            weight_name = wp.weight_name
            node_name = wp.node_with_weight.node_name
            config = wp.compression_config

            if config.num_bits != 4 or node_name not in self._statistics:
                scales[weight_name] = None
                continue

            stats = self._statistics[node_name]

            weight_data = self._backend_entity.get_weight_names_and_port_ids(wp.node_with_weight, graph)
            if len(weight_data) != 1:  # not supported by the algorithm
                continue
            _, weight_port_id = weight_data[0]

            weight = self._backend_entity.get_weight(wp.node_with_weight, weight_port_id, model, graph)

            scales[weight_name], zero_points[weight_name] = self.calculate_quantization_params(
                stats,
                weight,
                wp.reduction_axes,
                config,
                self._subset_size,
                self._initial_steps,
                self._scale_steps,
                self._weight_penalty,
            )

        return scales, zero_points

    @staticmethod
    def calculate_quantization_params(
        statistics: WCTensorStatistic,
        weight: Tensor,
        reduction_axes: Tuple[int, ...],
        config: WeightCompressionConfig,
        subset_size: int = 32,
        initial_steps: int = 5,
        scale_steps: int = 10,
        weight_penalty: float = -1.0,
    ) -> Tensor:
        """
        Calculates the quantization parameters for a given set of weights and activations.
        This function estimates the optimal quantization scale for weight compression by
        minimizing the difference between floating-point operations and operations with
        quantized weights.

        The function uses an iterative process:
        1. Initial scale rectification based on activation statistics.
        2. A grid search to further refine the scale parameters.

        :param statistics: The input activations of the layer reduced over batch and sequence length dimensions,
            together with original activation tensor shapes.
        :param weight: The weight tensor that is being quantized.
        :param reduction_axes: Tuple specifying the axes along which the reduction is performed for quantization.
        :param config: Configuration parameters for the weight compression, including quantization settings.
        :param subset_size: The number of samples to use for scale estimation. Defaults to 32.
        :param initial_steps: The number of steps for initial scale rectification using activation statistics.
            Defaults to 5.
        :param scale_steps: The number of steps for refining the scale using a grid search. Defaults to 10.
        :param weight_penalty: Penalty coefficient applied to the difference between floating-point
            and quantized weights. A value of -1 disables the penalty. Defaults to -1.0.
        :return: A tensor containing the calculated quantization scales and zero points if applicable.
        """
        reduction_axis = reduction_axes[0]

        s, X = process_stats(statistics, subset_size)

        weight = weight.astype(TensorDataType.float32)
        eps = fns.finfo(weight).eps

        if reduction_axis == 0:
            weight = fns.transpose(weight)
            reduction_axis = 1

        group_size = config.group_size if config.group_size != -1 else weight.shape[reduction_axis]
        cur_config = deepcopy(config)
        cur_config.group_size = group_size

        original_weight = fns.zeros_like(weight) + weight
        if config.mode == CompressWeightsMode.NF4:
            norm_weight, scale = calculate_normalized_weight_and_fp4_scale(
                original_weight, reduction_axis, cur_config.group_size
            )
            compressed_weights = do_nf4_quantization(norm_weight, scale, is_normalized_weight=True)
            q_weights = do_nf4_dequantization(compressed_weights, scale, reduction_axis)
            q_weights, _ = reshape_weight_for_grouped_quantization(q_weights, reduction_axis, group_size)
            zp = None
        else:
            q_weights, compressed_weights, scale, zp = quantize_dequantize_weight(
                original_weight, cur_config, reduction_axis, return_compressed_weight=True
            )
            if zp is not None:
                zp = zp.astype(scale.dtype)

        s = fns.unsqueeze(s, 0)
        s, _ = reshape_weight_for_grouped_quantization(s, reduction_axis, group_size)

        original_weight, _ = reshape_weight_for_grouped_quantization(original_weight, reduction_axis, group_size)

        # all weight in group has importance based on corresponding input activations
        importance = fns.ones_like(original_weight)
        importance = importance * s

        target, zero_mask = get_target_zero_mask(compressed_weights, zp)
        importance = fns.where(zero_mask, 0.0, importance)

        # normalize importances for every group of weights to make sum of them equal to 1.0
        denum = fns.sum(importance, axis=2, keepdims=True)
        importance = importance / (denum + eps)

        X, _ = reshape_weight_for_grouped_quantization(X, 0, group_size)
        best_diffs = None
        result_scale = None

        fp_outs = fns.matmul(fns.transpose(original_weight, (1, 0, 2)), X)
        q_outs = fns.matmul(fns.transpose(q_weights, (1, 0, 2)), X)

        # metric for minimization with shape [C_OUT, N_GROUPS], N_GROUPS = C_IN / GROUP_SIZE
        min_max_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
        min_max_scale_diffs = fns.transpose(min_max_scale_diffs, (1, 0))
        if weight_penalty > 0.0:
            min_max_scale_diffs += weight_penalty * fns.mean((q_weights - original_weight) ** 2, axis=-1)

        scale_sign = scale / fns.abs(scale)
        zero_scale = 0.001
        zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)

        # This is required for alignment with a previous OpenVINO models implementation
        # TODO(Nikita Savelyev): remove this
        opt_fns_kwargs = dict(dynamic_shapes=False, convertable_division=True)

        # iterative rectification of initial scale
        for i in range(initial_steps):
            near_to_ideal_scale = estimate_scales(original_weight, target, zero_mask, importance)
            near_to_ideal_scale = near_to_ideal_scale * scale_sign

            if config.mode == CompressWeightsMode.NF4:
                g_compressed_weighs = do_nf4_quantization(original_weight, near_to_ideal_scale)
                out = do_nf4_dequantization(g_compressed_weighs, near_to_ideal_scale)
            else:
                out = quantize_dequantize_weight(
                    original_weight,
                    config,
                    precomputed_scale=near_to_ideal_scale,
                    precomputed_zero_point=zp,
                    **opt_fns_kwargs,
                )

            q_weights_ = fns.zeros_like(original_weight) + out
            q_outs = fns.matmul(fns.transpose(q_weights_, (1, 0, 2)), X)

            ideal_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
            ideal_scale_diffs = fns.transpose(ideal_scale_diffs, (1, 0))
            if weight_penalty > 0.0:
                ideal_scale_diffs += weight_penalty * fns.mean((q_weights_ - original_weight) ** 2, axis=-1)

            if best_diffs is None:
                best_diffs = min_max_scale_diffs

            mask = (ideal_scale_diffs > best_diffs).astype(best_diffs.dtype)

            best_diffs = mask * best_diffs + (1.0 - mask) * ideal_scale_diffs

            mask = fns.unsqueeze(mask, axis=2)

            if result_scale is None:
                near_to_ideal_scale = mask * scale + (1.0 - mask) * near_to_ideal_scale
            else:
                near_to_ideal_scale = mask * result_scale + (1.0 - mask) * near_to_ideal_scale
            result_scale = near_to_ideal_scale

            if i < initial_steps - 1:
                if config.mode == CompressWeightsMode.NF4:
                    out = do_nf4_quantization(original_weight, near_to_ideal_scale)
                else:
                    out, _, _ = do_int_quantization(
                        original_weight,
                        config,
                        precomputed_scale=near_to_ideal_scale,
                        precomputed_zero_point=zp,
                        **opt_fns_kwargs,
                    )
                compressed_weights = fns.zeros_like(original_weight) + out
                target, zero_mask = get_target_zero_mask(compressed_weights, zp)
                zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)

        # iterative rectification of scale based on grid search
        for scale_step in range(scale_steps):
            factor = 1.0 - 0.05 * scale_step
            scaled_scale = factor * scale

            if config.mode == CompressWeightsMode.NF4:
                out = do_nf4_quantization(original_weight, scaled_scale)
            else:
                out, _, _ = do_int_quantization(
                    original_weight,
                    config,
                    precomputed_scale=scaled_scale,
                    precomputed_zero_point=zp,
                    **opt_fns_kwargs,
                )
            compressed_weights = fns.zeros_like(original_weight) + out

            target, zero_mask = get_target_zero_mask(compressed_weights, zp)
            zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)
            near_to_ideal_scale = estimate_scales(original_weight, target, zero_mask, importance)
            near_to_ideal_scale = near_to_ideal_scale * scale_sign

            if config.mode == CompressWeightsMode.NF4:
                g_compressed_weighs = do_nf4_quantization(original_weight, near_to_ideal_scale)
                out = do_nf4_dequantization(g_compressed_weighs, near_to_ideal_scale)
            else:
                out = quantize_dequantize_weight(
                    original_weight,
                    config,
                    precomputed_scale=near_to_ideal_scale,
                    precomputed_zero_point=zp,
                    **opt_fns_kwargs,
                )
            q_weights_ = fns.zeros_like(original_weight) + out

            q_outs = fns.matmul(fns.transpose(q_weights_, (1, 0, 2)), X)
            ideal_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
            ideal_scale_diffs = fns.transpose(ideal_scale_diffs, (1, 0))
            if weight_penalty > 0.0:
                ideal_scale_diffs += weight_penalty * fns.mean((q_weights_ - original_weight) ** 2, axis=-1)

            mask = (ideal_scale_diffs > best_diffs).astype(best_diffs.dtype)

            best_diffs = mask * best_diffs + (1.0 - mask) * ideal_scale_diffs

            mask = fns.unsqueeze(mask, axis=2)

            if result_scale is None:
                near_to_ideal_scale = mask * scale + (1.0 - mask) * near_to_ideal_scale
            else:
                near_to_ideal_scale = mask * result_scale + (1.0 - mask) * near_to_ideal_scale
            result_scale = near_to_ideal_scale

        if config.group_size == -1:
            result_scale = fns.squeeze(result_scale, axis=1)
        if zp is not None and config.group_size == -1:
            zp = fns.squeeze(zp, axis=1)

        return result_scale, zp

    @staticmethod
    def activations_to_wc_statistics(activations: List[Tensor]) -> WCTensorStatistic:
        """
        Mimic the activation reducing logic from WeightCompression.get_statistic_points.

        :param activations: List of raw activations.
        :return: Instance of WCTensorStatistic class containing reduced activations and shapes.
        """
        mean_values = []
        shapes = []
        for act in activations:
            shapes.append(act.shape)
            reduction_shape = tuple(range(act.ndim - 1))
            mean_values.append(fns.mean(act, axis=reduction_shape))
        wc_statistics = WCTensorStatistic(mean_values, shapes)
        return wc_statistics


def get_target_zero_mask(compressed_weights: Tensor, zp: Optional[Tensor] = None) -> Tuple[Tensor, Tensor]:
    """
    Computes the target values and a mask indicating zero values in the target.

    :param compressed_weights: The compressed weights tensor.
    :param zp: The zero point tensor.
    :return: The compressed weights optionally adjusted by the zero point and
        a boolean mask indicating positions in the target that are close to zero.
    """
    target = compressed_weights
    if zp is not None:
        target = target.astype(dtype=zp.dtype) - zp
    zero_mask = fns.isclose(target, 0)
    return target, zero_mask


def estimate_scales(weight: Tensor, target: Tensor, zero_mask: Tensor, importance: Tensor) -> Tensor:
    """
    Estimates scales for the given weight, target, zero mask, and importance.

    :param weight: The weights tensor.
    :param target: The target values tensor.
    :param zero_mask: A boolean mask indicating positions in the target that are close to zero.
    :param importance: The importance values tensor.
    :return: The estimated scales
    """
    ideal_scale = fns.abs(weight) / (fns.abs(target) + zero_mask)
    weighted_scale = ideal_scale * importance
    near_to_ideal_scale = fns.sum(weighted_scale, axis=2, keepdims=True)
    return near_to_ideal_scale



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from enum import Enum
from typing import Dict, List, Set, Tuple

import torch

import nncf
from nncf.common.deprecation import warning_deprecated
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api


@api(canonical_alias="nncf.torch.load_state")
def load_state(
    model: torch.nn.Module, state_dict_to_load: dict, is_resume: bool = False, keys_to_ignore: List[str] = None
) -> int:
    """
    Used to load a checkpoint containing a compressed model into an NNCFNetwork object, but can
    be used for any PyTorch module as well. Will do matching of state_dict_to_load parameters to
    the model's state_dict parameters while discarding irrelevant prefixes added during wrapping
    in NNCFNetwork or DataParallel/DistributedDataParallel objects, and load the matched parameters
    from the state_dict_to_load into the model's state dict.
    :param model: The target module for the state_dict_to_load to be loaded to.
    :param state_dict_to_load: A state dict containing the parameters to be loaded into the model.
    :param is_resume: Determines the behavior when the function cannot do a successful parameter match
    when loading. If True, the function will raise an exception if it cannot match the state_dict_to_load
    parameters to the model's parameters (i.e. if some parameters required by model are missing in
    state_dict_to_load, or if state_dict_to_load has parameters that could not be matched to model parameters,
    or if the shape of parameters is not matching). If False, the exception won't be raised.
    Usually is_resume is specified as False when loading uncompressed model's weights into the model with
    compression algorithms already applied, and as True when loading a compressed model's weights into the model
    with compression algorithms applied to evaluate the model.
    :param keys_to_ignore: A list of parameter names that should be skipped from matching process.
    :return: The number of state_dict_to_load entries successfully matched and loaded into model.
    """

    model_state_dict = model.state_dict()

    from nncf.torch.utils import maybe_convert_legacy_names_in_model_state

    maybe_convert_legacy_names_in_model_state(state_dict_to_load)
    key_matcher = KeyMatcher(is_resume, state_dict_to_load, model_state_dict, keys_to_ignore)
    new_dict = key_matcher.run()
    num_loaded_params = len(new_dict)
    key_matcher.handle_problematic_keys()
    nncf_logger.info(f"Loaded {num_loaded_params}/{len(model_state_dict.items())} parameters")

    model.load_state_dict(new_dict, strict=False)
    return num_loaded_params


class ParametersRegistry:
    """
    Provides an interface to register parameters and get access to all of them.
    """

    def __init__(self):
        self._parameters_names = set()

    def register(self, parameter_name: str):
        self._parameters_names.add(parameter_name)

    def get_parameters_names(self) -> Set[str]:
        return self._parameters_names


# If optional parameter is missed in a checkpoint, it can be loaded without an error in a strict mode.
# New parameters can be introduced for the model without breaking backward compatibility with old checkpoint.
OPTIONAL_PARAMETERS_REGISTRY = ParametersRegistry()


class ProcessedKeyStatus(Enum):
    """Status of matching checkpoint key with model keys"""

    MATCHED = "Matched"
    MISSING = "Missing"
    UNEXPECTED = "Unexpected"
    SIZE_MISMATCHED = "Size mismatched"
    SKIPPED = "Skipped"


class ProcessedKeys:
    """Contains checkpoint keys with their status of matching with model keys"""

    def __init__(self):
        self._keys: Dict[ProcessedKeyStatus, Set[str]] = {}
        for key_status in ProcessedKeyStatus:
            self._keys[key_status] = set()

    def add_key(self, key: str, status: ProcessedKeyStatus):
        self._keys[status].add(key)

    def extend_keys(self, keys: List[str], status: ProcessedKeyStatus):
        self._keys[status].update(keys)

    def add_skipped_and_missing_keys(self, model_state_dict: Dict[str, torch.Tensor]):
        all_processed_keys = []
        optional_param_names = OPTIONAL_PARAMETERS_REGISTRY.get_parameters_names()
        params_to_skip = tuple("." + name for name in optional_param_names)
        for keys in self._keys.values():
            all_processed_keys.extend(keys)

        for key in model_state_dict:
            if key not in all_processed_keys:
                if key.endswith(params_to_skip) or key in optional_param_names:
                    self.add_key(key, ProcessedKeyStatus.SKIPPED)
                    nncf_logger.warning(f"The optional parameter {key} is missing in the loaded state.")
                else:
                    self.add_key(key, ProcessedKeyStatus.MISSING)

    def handle_problematic(self, is_resume: bool, are_all_loaded_params_matched: bool):
        """
        Reports about errors during the matching state_dict_to_load parameters to the model's state_dict ones.
        It raises an error if is_resume is True or prints warning when it's False. The report happens if
        state_dict_to_load has parameters that could not be matched to model parameters or if the shape of parameters is
        not matching. If some parameters required by model are missing in state_dict_to_load reporting occurs for
        non optional parameters only or when not all parameters from state_dict_to_load match.
        :param is_resume: Determines the behavior when the function cannot do a successful parameter match when loading.
        :param are_all_loaded_params_matched: whether all parameters to load match with model parameters
        """
        error_msgs = []

        def add_error_msg(name, keys_):
            error_msgs.insert(0, "{} key(s):\n{}. ".format(name, ",\n".join('\t\t"{}"'.format(k) for k in keys_)))

        for key_status, keys in self._keys.items():
            is_missing = key_status == ProcessedKeyStatus.MISSING
            erroneous = key_status in (ProcessedKeyStatus.SIZE_MISMATCHED, ProcessedKeyStatus.UNEXPECTED)
            if keys and (erroneous or is_missing and (is_resume or not are_all_loaded_params_matched)):
                add_error_msg(key_status.value, keys)
        if error_msgs:
            error_msg = "Error(s) when loading model parameters:\n\t{}".format("\n\t".join(error_msgs))
            if is_resume:
                raise nncf.InternalError(error_msg)
            nncf_logger.error(error_msg)


class NormalizedKeys:
    """
    Contains normalized form of parameters. It helps to discard irrelevant prefixes added during wrapping in
    NNCFNetwork or DataParallel/DistributedDataParallel objects, to handle legacy parameters' names and to match
    unified compression parameters from the separate ones.
    """

    def __init__(self, keys: List[str], keys_to_ignore: List[str]):
        self._unique_normalized_key_vs_orig_key_map = {}
        self.is_unified_group_detected = False
        unique_clipped_key_vs_orig_key_map, ignored_keys = self._clip_keys_without_collisions(keys, keys_to_ignore)
        self.ignored_orig_keys = ignored_keys
        ignored_keys = self._normalize_keys_without_collisions(unique_clipped_key_vs_orig_key_map, keys_to_ignore)
        self.ignored_orig_keys.extend(ignored_keys)

    def __contains__(self, key: str):
        return key in self._unique_normalized_key_vs_orig_key_map

    def __iter__(self):
        return iter(self._unique_normalized_key_vs_orig_key_map)

    def get_orig_key(self, normalized_key: str):
        return self._unique_normalized_key_vs_orig_key_map[normalized_key]

    def _normalize_keys_without_collisions(
        self, unique_clipped_key_vs_orig_key_map: Dict[str, str], keys_to_ignore: List[str]
    ) -> List[str]:
        ignored_keys = []
        normalized_key_vs_clipped_key_list_map = {}
        for clipped_key in unique_clipped_key_vs_orig_key_map:
            replaced_keys = self._key_replacer(clipped_key)
            if len(replaced_keys) > 1:
                self.is_unified_group_detected = True

            for replaced_key in replaced_keys:
                if replaced_key in keys_to_ignore:
                    orig_key = unique_clipped_key_vs_orig_key_map[clipped_key]
                    ignored_keys.append(orig_key)
                    continue
                if replaced_key in normalized_key_vs_clipped_key_list_map:
                    normalized_key_vs_clipped_key_list_map[replaced_key].append(clipped_key)
                else:
                    normalized_key_vs_clipped_key_list_map[replaced_key] = [clipped_key]
        # keep clipped keys if their normalization led to a collisions
        for normalized_key, value in normalized_key_vs_clipped_key_list_map.items():
            list_clipped_keys = value
            if len(list_clipped_keys) == 1:
                clipped_key = list_clipped_keys[0]
                orig_key = unique_clipped_key_vs_orig_key_map[clipped_key]
                self._unique_normalized_key_vs_orig_key_map[normalized_key] = orig_key
            else:
                for clipped_key in list_clipped_keys:
                    orig_key = unique_clipped_key_vs_orig_key_map[clipped_key]
                    self._unique_normalized_key_vs_orig_key_map[clipped_key] = orig_key
        return ignored_keys

    @staticmethod
    def _clip_keys_without_collisions(keys: List[str], keys_to_ignore: List[str]) -> Tuple[Dict[str, str], List[str]]:
        clipped_key_vs_orig_key_list_map = {}
        ignored_keys = []
        for orig_key in keys:
            clipped_key = NormalizedKeys._key_clipper(orig_key)
            if clipped_key in keys_to_ignore:
                ignored_keys.append(orig_key)
                continue
            if clipped_key in clipped_key_vs_orig_key_list_map:
                clipped_key_vs_orig_key_list_map[clipped_key].append(orig_key)
            else:
                clipped_key_vs_orig_key_list_map[clipped_key] = [orig_key]
        # keep original keys if their clipping led to a collisions
        unique_clipped_key_vs_orig_key_map = {}
        for clipped_key, val in clipped_key_vs_orig_key_list_map.items():
            list_orig_keys = val
            if len(list_orig_keys) == 1:
                unique_clipped_key_vs_orig_key_map[clipped_key] = list_orig_keys[0]
            else:
                for orig_key in list_orig_keys:
                    unique_clipped_key_vs_orig_key_map[orig_key] = orig_key
        return unique_clipped_key_vs_orig_key_map, ignored_keys

    @staticmethod
    def _key_clipper(key: str) -> str:
        new_key = key

        clip_patterns = ["module.", "|OUTPUT", "|INPUT"]
        for pattern in clip_patterns:
            new_key = new_key.replace(pattern, "")
        return new_key

    def _key_replacer(self, key: str) -> List[str]:
        new_key = key

        match = re.search("(pre_ops|post_ops)\\.(\\d+?)\\.op", key)
        new_key = new_key if not match else new_key.replace(match.group(), "operation")
        result = self._split_unified_parameters(new_key)
        if len(result) > 1:
            self.is_unified_group_detected = True
        return result

    @staticmethod
    def _split_unified_parameters(new_key: str) -> List[str]:
        """covers unified activation quantizers case, e.g.
            external_quantizers.RELU_0;RELU_1;RELU_2.op
        Result of this function is full names of individual parameters:
            external_quantizers.RELU_2.op
            external_quantizers.RELU_1.op
            external_quantizers.RELU_0.op
        It's utilized to match parameters from checkpoints without unified operations to not start training
        compression from scratch, but instead initialize group of parameters by one of the matched individual one.
        Returns original key if there's no ';' and operation doesn't start with EXTERNAL_QUANTIZERS_STORAGE_NAME
        """
        result = [new_key]
        from nncf.torch.quantization.external_quantizer import EXTERNAL_QUANTIZERS_STORAGE_PREFIX

        if ";" in new_key and new_key.startswith(EXTERNAL_QUANTIZERS_STORAGE_PREFIX):
            group_of_keys = new_key.split(";")
            last_key = group_of_keys[-1]
            common_op = last_key.split(".")[-1]
            result = [group_of_keys[0] + "." + common_op, EXTERNAL_QUANTIZERS_STORAGE_PREFIX + "." + last_key]
            for key in group_of_keys[1:-1]:
                result.append(EXTERNAL_QUANTIZERS_STORAGE_PREFIX + "." + key + "." + common_op)
        return result


class KeyMatcher:
    """
    Matches state_dict_to_load parameters to the model's state_dict parameters while discarding irrelevant prefixes
    added during wrapping in NNCFNetwork or DataParallel/DistributedDataParallel objects, skipping registered optional
    parameters, handling legacy parameters' names, ignoring the order of pre/post operations, matching unified
    compression parameters from the separate ones and forms the model state dict with matched parameters.
    """

    def __init__(
        self,
        is_resume: bool,
        state_dict_to_load: Dict[str, torch.Tensor],
        model_state_dict: Dict[str, torch.Tensor],
        ignored_keys: List[str] = None,
    ):
        """
        :param state_dict_to_load: A state dict containing the parameters to be loaded into the model.
        :param ignored_keys: list of parameters to skip from matching process on loading.
        """
        self._is_resume = is_resume
        self.state_dict_to_load = state_dict_to_load

        self.model_state_dict = model_state_dict
        self._processed_keys = ProcessedKeys()
        self._new_dict = {}
        self._num_params_to_load = len(state_dict_to_load.items())
        self.ignored_keys = ignored_keys if ignored_keys else []

    def run(self) -> Dict[str, torch.Tensor]:
        """
        :return: the model state dict with matched parameters
        """
        normalized_model_keys = NormalizedKeys(list(self.model_state_dict.keys()), keys_to_ignore=self.ignored_keys)
        normalized_keys_to_load = NormalizedKeys(list(self.state_dict_to_load.keys()), keys_to_ignore=self.ignored_keys)

        has_version_agnostic_names = False
        cross_match_key_map = self._cross_match_version_agnostic_names(
            list(normalized_keys_to_load), list(normalized_model_keys)
        )

        for matched_checkpoint_key, matched_model_key in cross_match_key_map.items():
            if matched_checkpoint_key != matched_model_key:
                has_version_agnostic_names = True

        if has_version_agnostic_names:
            warning_deprecated(
                "Legacy NNCF-enabled .pth checkpoint has been loaded! "
                "The version-agnostic `RELU` operator name entries in the state dict "
                "have been deprecated. "
                "The loader will try to match these entries to the corresponding `relu` and `relu_` op "
                "names. The newly exported checkpoints will be adjusted to the new format."
            )

        if normalized_model_keys.is_unified_group_detected and not normalized_keys_to_load.is_unified_group_detected:
            nncf_logger.warning(
                "Unified parameters are detected in the compressed model, but all parameters are independent "
                "and separate in the loading checkpoint. The unified parameters will be initialized by one of"
                "the corresponding separate parameter in the checkpoint. That may slightly degrade the "
                "accuracy, but should allow to not start training compression from scratch with unified "
                "params."
            )
        ignored_keys = normalized_model_keys.ignored_orig_keys + normalized_keys_to_load.ignored_orig_keys
        self._processed_keys.extend_keys(ignored_keys, ProcessedKeyStatus.SKIPPED)
        if ignored_keys:
            ignored_keys_str = "\n".join(set(ignored_keys))
            nncf_logger.warning(
                f"Following parameters were skipped from matching checkpoint's keys:\n{ignored_keys_str}"
            )

        loaded_prefixless_keys = False
        for normalized_key_to_load in normalized_keys_to_load:
            key_to_load = normalized_keys_to_load.get_orig_key(normalized_key_to_load)
            normalized_key_to_load = cross_match_key_map.get(normalized_key_to_load, normalized_key_to_load)
            if normalized_key_to_load in normalized_model_keys:
                model_key = normalized_model_keys.get_orig_key(normalized_key_to_load)
                if "_nncf." + key_to_load == model_key:
                    loaded_prefixless_keys = True
                value_to_load = self.state_dict_to_load[key_to_load]
                size_of_value_to_load = value_to_load.size()
                size_of_model_value = self.model_state_dict[model_key].size()
                if size_of_value_to_load == size_of_model_value:
                    self._new_dict[model_key] = value_to_load
                    self._processed_keys.add_key(model_key, ProcessedKeyStatus.MATCHED)
                else:
                    nncf_logger.warning(
                        f"Different size of value of '{model_key}' "
                        f"in resuming dictionary ({size_of_value_to_load}) and in model ({size_of_model_value})"
                    )
                    self._processed_keys.add_key(model_key, ProcessedKeyStatus.SIZE_MISMATCHED)
            else:
                self._processed_keys.add_key(key_to_load, ProcessedKeyStatus.UNEXPECTED)
        self._processed_keys.add_skipped_and_missing_keys(self.model_state_dict)
        if loaded_prefixless_keys:
            warning_deprecated(
                "Legacy NNCF-enabled .pth checkpoint has been loaded! "
                'Some storage keys in the loaded checkpoint should now have a "_nncf." prefix,'
                "support for the legacy storage key will be dropped in a future release. "
                "This checkpoint will be loaded; update your checkpoint file by saving this model's"
                "checkpoint file again."
            )
        return self._new_dict

    @staticmethod
    def _cross_match_version_agnostic_names(
        normalized_keys_to_load: List[str], normalized_model_keys: List[str]
    ) -> Dict[str, str]:
        """
        Handles the situation where the normalized_keys_to_load contain legacy version-agnostic names
        of operations, such as `RELU`.

        :param normalized_keys_to_load: A list of keys in the checkpoint, potentially with version-agnostic names
        :param normalized_model_keys: A list of keys in the model, without version-agnostic names.
        :return: A mapping of the checkpoint key to a model key that matches version-agnostic names with their
            torch-specific counterparts.
        """
        version_agnostic_to_specific_names = {"RELU": {"relu", "relu_"}}
        retval = {}
        processed_keys_to_load = normalized_keys_to_load

        for model_key in normalized_model_keys:
            for agnostic_op_name, specific_op_name_set in version_agnostic_to_specific_names.items():
                matches_for_curr_agnostic_op_name = []
                has_specific_op_name = False
                for specific_op_name in specific_op_name_set:
                    # Have to take care not to replace the matches to the class names
                    # The op names in existing checkpoint can only appear in external quantizers,
                    # i.e. external_quantizers.ResNet/ReLU[relu]/relu_0.signed_tensor, so composing a regex to match
                    # for that
                    slash_split_str = model_key.split("/")
                    last_portion = slash_split_str[-1]
                    if specific_op_name in last_portion:
                        last_portion = last_portion.replace(specific_op_name, agnostic_op_name, 1)
                        has_specific_op_name = True
                    slash_split_str[-1] = last_portion
                    agnostic_version_of_model_key = "/".join(slash_split_str)
                    processed_agnostic_version_of_model_key = agnostic_version_of_model_key
                    if processed_agnostic_version_of_model_key in processed_keys_to_load:
                        idx = processed_keys_to_load.index(processed_agnostic_version_of_model_key)
                        matches_for_curr_agnostic_op_name.append(normalized_keys_to_load[idx])

                if not has_specific_op_name:
                    if matches_for_curr_agnostic_op_name:
                        checkpoint_matched_key = next(iter(matches_for_curr_agnostic_op_name))
                        retval[checkpoint_matched_key] = model_key
                elif len(matches_for_curr_agnostic_op_name) == 1:
                    checkpoint_matched_key = next(iter(matches_for_curr_agnostic_op_name))
                    retval[checkpoint_matched_key] = model_key
                elif len(matches_for_curr_agnostic_op_name) == 0:
                    nncf_logger.debug(f"Failed to match a version-specific key: {model_key}")
                elif len(matches_for_curr_agnostic_op_name) > 1:
                    nncf_logger.debug(
                        f"More than one match for the version specific key: {model_key}\n"
                        f"Matches:\n"
                        f"{', '.join(matches_for_curr_agnostic_op_name)}"
                    )

        return retval

    def handle_problematic_keys(self):
        """
        Reports about errors during the matching state_dict_to_load parameters to the model's state_dict ones.
        """
        num_matched_params = len(self._new_dict)
        self._processed_keys.handle_problematic(self._is_resume, num_matched_params == self._num_params_to_load)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Tuple

import torch


class NodeType(Enum):
    const = "const"
    fn_call = "function_call"
    input = "input"
    output = "output"

    def __str__(self) -> str:
        return self.value


class TensorSource(Enum):
    buffer = "buffer"
    function = "function"
    input = "input"
    output = "output"
    parameter = "parameter"

    def __str__(self) -> str:
        return self.value


@dataclass
class TensorMeta:
    dtype: torch.dtype
    shape: Tuple[int, ...]
    requires_grad: bool

    @staticmethod
    def from_tensor(tensor: torch.Tensor) -> TensorMeta:
        return TensorMeta(tensor.dtype, tuple(tensor.shape), tensor.requires_grad)


@dataclass
class ConstMeta:
    dtype: torch.dtype
    shape: Tuple[int, ...]
    name_in_model: str

    @staticmethod
    def from_tensor(tensor: torch.Tensor, name_in_model: str) -> ConstMeta:
        return ConstMeta(tensor.dtype, tuple(tensor.shape), name_in_model)


@dataclass
class InOutMeta:
    dtype: torch.dtype
    shape: Tuple[int, ...]
    name: str

    @staticmethod
    def from_tensor(tensor: torch.Tensor, name: str) -> InOutMeta:
        return InOutMeta(tensor.dtype, tuple(tensor.shape), name)


@dataclass
class FunctionMeta:
    op_name: str
    fn_name: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


@dataclass
class EdgeMeta:
    dtype: torch.dtype
    shape: Tuple[int, ...]
    input_port: int
    output_port: int

    @staticmethod
    def from_tensor(tensor: torch.Tensor, input_port: int, output_port: int) -> EdgeMeta:
        return EdgeMeta(tensor.dtype, tuple(tensor.shape), input_port, output_port)


@dataclass
class TensorInfo:
    tensor_source: TensorSource
    shape: Tuple[int, ...]
    dtype: torch.dtype
    output_port_id: int
    source_node_id: Optional[int]
    name_in_model: Optional[str]
